"""
Module for obtaining data from Amazon Web Service
"""

import logging
import os
import warnings
from abc import ABC, abstractmethod

from .config import SHConfig
from .constants import AwsConstants, EsaSafeType, MimeType, DataSource
from .download import DownloadRequest
from .download.aws_client import get_aws_json
from .exceptions import AwsDownloadFailedException, SHUserWarning
from .opensearch import get_tile_info, get_tile_info_id
from .time_utils import parse_time


LOGGER = logging.getLogger(__name__)

MAX_SUPPORTED_BASELINES = {
    DataSource.SENTINEL2_L1C: '02.07',
    DataSource.SENTINEL2_L2A: '02.11'
}


class AwsService(ABC):
    """ Amazon Web Service (AWS) base class
    """
    def __init__(self, parent_folder='', bands=None, metafiles=None):
        """
        :param parent_folder: Folder where the fetched data will be saved.
        :type parent_folder: str
        :param bands: List of Sentinel-2 bands for request. If parameter is set to `None` all bands will be used.
        :type bands: list(str) or None
        :param metafiles: List of additional metafiles available on AWS
                          (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``).
                          If parameter is set to `None` the list will be set automatically.
        :type metafiles: list(str) or None
        """
        self.parent_folder = parent_folder
        self.bands = self._parse_bands(bands)
        self.metafiles = self._parse_metafiles(metafiles)

        self.download_list = []
        self.folder_list = []

        self.base_url = self.get_base_url()
        self.base_http_url = self.get_base_url(force_http=True)

    @abstractmethod
    def get_requests(self):
        """ Abstract class for joining together download requests
        """
        raise NotImplementedError

    def _parse_bands(self, band_input):
        """
        Parses class input and verifies band names.

        :param band_input: input parameter `bands`
        :type band_input: str or list(str) or None
        :return: verified list of bands
        :rtype: list(str)
        """
        all_bands = AwsConstants.S2_L1C_BANDS if self.data_source is DataSource.SENTINEL2_L1C else \
            AwsConstants.S2_L2A_BANDS

        if band_input is None:
            return all_bands
        if isinstance(band_input, str):
            band_list = band_input.split(',')
        elif isinstance(band_input, list):
            band_list = band_input.copy()
        else:
            raise ValueError('bands parameter must be a list or a string')
        band_list = [band.strip().split('.')[0] for band in band_list]
        band_list = [band for band in band_list if band != '']
        if not set(band_list) <= set(all_bands):
            raise ValueError('bands {} must be a subset of {}'.format(band_list, all_bands))
        return band_list

    def _parse_metafiles(self, metafile_input):
        """ Parses class input and verifies metadata file names.

        :param metafile_input: class input parameter `metafiles`
        :type metafile_input: str or list(str) or None
        :return: verified list of metadata files
        :rtype: list(str)
        """
        all_metafiles = AwsConstants.S2_L1C_METAFILES if self.data_source is DataSource.SENTINEL2_L1C else \
            AwsConstants.S2_L2A_METAFILES

        if metafile_input is None:
            if self.__class__.__name__ == 'SafeProduct':
                return all_metafiles
            if self.__class__.__name__ == 'SafeTile':
                return [metafile for metafile in all_metafiles if metafile in AwsConstants.TILE_FILES]
            return []
        if isinstance(metafile_input, str):
            metafile_list = metafile_input.split(',')
        elif isinstance(metafile_input, list):
            metafile_list = metafile_input.copy()
        else:
            raise ValueError('metafiles parameter must be a list or a string')
        metafile_list = [metafile.strip().split('.')[0] for metafile in metafile_list]
        metafile_list = [metafile for metafile in metafile_list if metafile != '']
        if not set(metafile_list) <= set(all_metafiles):
            raise ValueError('metadata files {} must be a subset of {}'.format(metafile_list, all_metafiles))
        return metafile_list

    def get_base_url(self, force_http=False):
        """ Creates base URL path

        :param force_http: `True` if HTTP base URL should be used and `False` otherwise
        :type force_http: str
        :return: base url string
        :rtype: str
        """
        base_url = SHConfig().aws_metadata_url if force_http else 's3:/'
        aws_bucket = SHConfig().aws_s3_l1c_bucket if self.data_source is DataSource.SENTINEL2_L1C else \
            SHConfig().aws_s3_l2a_bucket

        return '{}/{}'.format(base_url, aws_bucket)

    def get_safe_type(self):
        """ Determines the type of ESA product.

        In 2016 ESA changed structure and naming of data. Therefore the class must
        distinguish between old product type and compact (new) product type.

        :return: type of ESA product
        :rtype: constants.EsaSafeType
        :raises: ValueError
        """
        product_type = self.product_id.split('_')[1]
        if product_type.startswith('MSI'):
            return EsaSafeType.COMPACT_TYPE
        if product_type in ['OPER', 'USER']:
            return EsaSafeType.OLD_TYPE
        raise ValueError('Unrecognized product type of product id {}'.format(self.product_id))

    def get_baseline(self):
        """ Determines the baseline number (i.e. version) of ESA .SAFE product

        :return: baseline number
        :rtype: str
        :raises: ValueError
        """
        if self.safe_type is EsaSafeType.COMPACT_TYPE:
            baseline = self.product_id.split('_')[3].lstrip('N')
            if len(baseline) != 4:
                raise ValueError('Unable to recognize baseline number from the product id {}'.format(self.product_id))
            baseline = '{}.{}'.format(baseline[:2], baseline[2:])

            if baseline > MAX_SUPPORTED_BASELINES[self.data_source]:
                message = 'Products with baseline {} are not officially supported in sentinelhub-py. If you notice ' \
                          'any errors in naming structure of downloaded data please report an issue at ' \
                          'https://github.com/sentinel-hub/sentinelhub-py/issues. Pull requests are also very ' \
                          'appreciated'.format(baseline)
                warnings.warn(message, category=SHUserWarning)

            return baseline
        return self._read_baseline_from_info()

    def _read_baseline_from_info(self):
        """ Tries to find and return baseline number from either tileInfo or productInfo file.

        :return: Baseline ID
        :rtype: str
        :raises: ValueError
        """
        if hasattr(self, 'tile_info'):
            return self.tile_info['datastrip']['id'][-5:]
        if hasattr(self, 'product_info'):
            return self.product_info['datastrips'][0]['id'][-5:]
        raise ValueError('No info file has been obtained yet.')

    @staticmethod
    def url_to_tile(url):
        """ Extracts tile name, date and AWS index from tile url on AWS.

        :param url: class input parameter 'metafiles'
        :type url: str
        :return: Name of tile, date and AWS index which uniquely identifies tile on AWS
        :rtype: (str, str, int)
        """
        info = url.strip('/').split('/')
        name = ''.join(info[-7: -4])
        date = '-'.join(info[-4: -1])
        return name, date, int(info[-1])

    def sort_download_list(self):
        """ Method for sorting the list of download requests. Band images have priority before metadata files. If bands
        images or metadata files are specified with a list they will be sorted in the same order as in the list.
        Otherwise they will be sorted alphabetically (band B8A will be between B08 and B09).
        """
        def aws_sort_function(download_request):
            data_name = download_request.properties['data_name']
            if 'product_name' in download_request.properties:
                product_name = download_request.properties['product_name']
            else:
                product_name = self._url_to_props(download_request.url)[0]
            if data_name in self.bands:
                return 0, product_name, self.bands.index(data_name)
            return 1, product_name, self.metafiles.index(data_name)
        self.download_list.sort(key=aws_sort_function)

    def structure_recursion(self, struct, folder):
        """ From nested dictionaries representing .SAFE structure it recursively extracts all the files that need to be
        downloaded and stores them into class attribute `download_list`.

        :param struct: nested dictionaries representing a part of .SAFE structure
        :type struct: dict
        :param folder: name of folder where this structure will be saved
        :type folder: str
        """
        has_subfolder = False
        for name, substruct in struct.items():
            subfolder = os.path.join(folder, name)
            if not isinstance(substruct, dict):
                product_name, data_name = self._url_to_props(substruct)
                if '.' in data_name:
                    data_type = MimeType(data_name.split('.')[-1])
                    data_name = data_name.rsplit('.', 1)[0]
                else:
                    data_type = MimeType.RAW
                if data_name in self.bands + self.metafiles:
                    self.download_list.append(DownloadRequest(url=substruct, filename=subfolder, data_type=data_type,
                                                              data_name=data_name, product_name=product_name))
            else:
                has_subfolder = True
                self.structure_recursion(substruct, subfolder)
        if not has_subfolder:
            self.folder_list.append(folder)

    def _url_to_props(self, url):
        """ Converts url back to name of product/tile and name of the file

        :param url: URL location of the data
        :type url: str
        :return: Names of product or tile and name of a file
        :rtype: str, str
        """
        props = (url[len(self.base_url):] if url.startswith(self.base_url) else
                 url[len(self.base_http_url):]).lstrip('/').split('/')
        if props[0] == 'products':
            tile_props = props[:5]
            props = props[5:]
            if props[0] == 'datastrip':
                props[1] = '*'
        else:
            tile_props = props[:8]
            props = props[8:]
        return '/'.join(tile_props), '/'.join(props)

    @staticmethod
    def add_file_extension(filename, data_format=None, remove_path=False):
        """ Joins filename and corresponding file extension if it has one.

        :param filename: Name of the file without extension
        :type filename: str
        :param data_format: format of file, if `None` it will be set automatically
        :type data_format: constants.MimeType or None
        :param remove_path: `True` if the path in filename string should be removed
        :type remove_path: bool
        :return: Name of the file with extension
        :rtype: str
        """
        if data_format is None:
            data_format = AwsConstants.AWS_FILES[filename]
        if remove_path:
            filename = filename.split('/')[-1]
        if filename.startswith('datastrip'):
            filename = filename.replace('*', '0')
        if data_format is MimeType.RAW:
            return filename
        return '{}.{}'.format(filename.replace('*', ''), data_format.value)

    def has_reports(self):
        """ Products created with baseline 2.06 and greater (and some products with baseline 2.05) should have quality
        report files

        :return: `True` if the product has report xml files and `False` otherwise
        :rtype: bool
        """
        return self.baseline > '02.05' or (self.baseline == '02.05' and self.date >= '2017-10-12')  # Not sure

    def is_early_compact_l2a(self):
        """ Check if product is early version of compact L2A product

        :return: `True` if product is early version of compact L2A product and `False` otherwise
        :rtype: bool
        """
        return self.data_source is DataSource.SENTINEL2_L2A and self.safe_type is EsaSafeType.COMPACT_TYPE and \
            self.baseline <= '02.06'


class AwsProduct(AwsService):
    """ Service class for Sentinel-2 product on AWS
    """
    def __init__(self, product_id, tile_list=None, **kwargs):
        """
        :param product_id: ESA ID of the product
        :type product_id: str
        :param tile_list: list of tile names
        :type tile_list: list(str) or None
        :param parent_folder: location of the directory where the fetched data will be saved.
        :type parent_folder: str
        :param bands: List of Sentinel-2 bands for request. If parameter is set to `None` all bands will be used.
        :type bands: list(str) or None
        :param metafiles: List of additional metafiles available on AWS
                          (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``).
                          If parameter is set to `None` the list will be set automatically.
        :type metafiles: list(str) or None
        """
        self.product_id = product_id.split('.')[0]
        self.tile_list = self.parse_tile_list(tile_list)

        self.data_source = self.get_data_source()
        self.safe_type = self.get_safe_type()

        super().__init__(**kwargs)

        self.date = self.get_date()
        self.product_url = self.get_product_url()
        self.product_info = get_aws_json(self.get_url(AwsConstants.PRODUCT_INFO))

        self.baseline = self.get_baseline()

    @staticmethod
    def parse_tile_list(tile_input):
        """ Parses class input and verifies band names.

        :param tile_input: class input parameter `tile_list`
        :type tile_input: str or list(str)
        :return: parsed list of tiles
        :rtype: list(str) or None
        """
        if tile_input is None:
            return None
        if isinstance(tile_input, str):
            tile_list = tile_input.split(',')
        elif isinstance(tile_input, list):
            tile_list = tile_input.copy()
        else:
            raise ValueError('tile_list parameter must be a list of tile names')
        tile_list = [AwsTile.parse_tile_name(tile_name) for tile_name in tile_list]
        return tile_list

    def get_requests(self):
        """ Creates product structure and returns list of files for download.

        :return: List of download requests and list of empty folders that need to be created
        :rtype: (list(download.DownloadRequest), list(str))
        """
        self.download_list = [DownloadRequest(url=self.get_url(metafile), filename=self.get_filepath(metafile),
                                              data_type=AwsConstants.AWS_FILES[metafile], data_name=metafile) for
                              metafile in self.metafiles if metafile in AwsConstants.PRODUCT_FILES]

        tile_parent_folder = os.path.join(self.parent_folder, self.product_id)
        for tile_info in self.product_info['tiles']:
            tile_name, date, aws_index = self.url_to_tile(self.get_tile_url(tile_info))
            if self.tile_list is None or AwsTile.parse_tile_name(tile_name) in self.tile_list:
                tile_downloads, tile_folders = AwsTile(tile_name, date, aws_index, parent_folder=tile_parent_folder,
                                                       bands=self.bands, metafiles=self.metafiles,
                                                       data_source=self.data_source).get_requests()
                self.download_list.extend(tile_downloads)
                self.folder_list.extend(tile_folders)
        self.sort_download_list()
        return self.download_list, self.folder_list

    def get_data_source(self):
        """ The method determines data source from product ID.

        :return: Data source of the product
        :rtype: DataSource
        :raises: ValueError
        """
        product_type = self.product_id.split('_')[1]
        if product_type.endswith('L1C') or product_type == 'OPER':
            return DataSource.SENTINEL2_L1C
        if product_type.endswith('L2A') or product_type == 'USER':
            return DataSource.SENTINEL2_L2A
        raise ValueError('Unknown data source of product {}'.format(self.product_id))

    def get_date(self):
        """ Collects sensing date of the product.

        :return: Sensing date
        :rtype: str
        """
        if self.safe_type == EsaSafeType.OLD_TYPE:
            name = self.product_id.split('_')[-2]
            date = [name[1:5], name[5:7], name[7:9]]
        else:
            name = self.product_id.split('_')[2]
            date = [name[:4], name[4:6], name[6:8]]
        return '-'.join(date_part.lstrip('0') for date_part in date)

    def get_url(self, filename, data_format=None):
        """ Creates url of file location on AWS.

        :param filename: name of file
        :type filename: str
        :param data_format: format of file, if `None` it will be set automatically
        :type data_format: constants.MimeType or None
        :return: url of file location
        :rtype: str
        """
        product_url = self.product_url
        force_http = filename in [AwsConstants.PRODUCT_INFO, AwsConstants.METADATA]
        if product_url is None or force_http:
            product_url = self.get_product_url(force_http=force_http)
        return '{}/{}'.format(product_url, self.add_file_extension(filename, data_format))

    def get_product_url(self, force_http=False):
        """ Creates base url of product location on AWS.

        :param force_http: `True` if HTTP base URL should be used and `False` otherwise
        :type force_http: str
        :return: url of product location
        :rtype: str
        """
        base_url = self.base_http_url if force_http else self.base_url
        return '{}/products/{}/{}'.format(base_url, self.date.replace('-', '/'), self.product_id)

    def get_tile_url(self, tile_info):
        """ Collects tile url from `productInfo.json` file.

        :param tile_info: information about tile from `productInfo.json`
        :type tile_info: dict
        :return: url of tile location
        :rtype: str
        """
        return '{}/{}'.format(self.base_url, tile_info['path'])

    def get_filepath(self, filename):
        """ Creates file path for the file.

        :param filename: name of the file
        :type filename: str
        :return: filename with path on disk
        :rtype: str
        """
        return os.path.join(self.parent_folder, self.product_id, self.add_file_extension(filename)).replace(':', '.')


class AwsTile(AwsService):
    """ Service class for Sentinel-2 product on AWS
    """
    def __init__(self, tile_name, time, aws_index=None, data_source=DataSource.SENTINEL2_L1C, **kwargs):
        """
        :param tile: Tile name (e.g. 'T10UEV')
        :type tile: str
        :param time: Tile sensing time in ISO8601 format
        :type time: str
        :param aws_index: There exist Sentinel-2 tiles with the same tile and time parameter. Therefore each tile on AWS
            also has an index which is visible in their url path. If ``aws_index`` is set to `None` the
            class will try to find the index automatically. If there will be multiple choices it will choose
            the lowest index and inform the user.
        :type aws_index: int or None
        :param data_source: Source of requested AWS data. Supported sources are Sentinel-2 L1C and Sentinel-2 L2A,
            default is Sentinel-2 L1C data.
        :type data_source: constants.DataSource
        :param parent_folder: folder where the fetched data will be saved.
        :type parent_folder: str
        :param bands: List of Sentinel-2 bands for request. If parameter is set to `None` all bands will be used.
        :type bands: list(str) or None
        :param metafiles: List of additional metafiles available on AWS
                          (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``).
                          If parameter is set to `None` the list will be set automatically.
        :type metafiles: list(str) or None
        """
        self.tile_name = self.parse_tile_name(tile_name)
        self.datetime = self.parse_datetime(time)
        self.date = self.datetime.split('T')[0]
        self.aws_index = aws_index
        self.data_source = data_source

        super().__init__(**kwargs)
        self.tile_url = None

        self.aws_index = self.get_aws_index()
        self.tile_url = self.get_tile_url()
        self.tile_info = self.get_tile_info()
        if not self.tile_is_valid():
            raise ValueError('Cannot find data on AWS for specified tile, time and aws_index')

        self.product_id = self.get_product_id()
        self.safe_type = self.get_safe_type()
        self.baseline = self.get_baseline()

    @staticmethod
    def parse_tile_name(name):
        """
        Parses and verifies tile name.

        :param name: class input parameter `tile_name`
        :type name: str
        :return: parsed tile name
        :rtype: str
        """
        tile_name = name.lstrip('T0')
        if len(tile_name) == 4:
            tile_name = '0' + tile_name
        if len(tile_name) != 5:
            raise ValueError('Invalid tile name {}'.format(name))
        return tile_name

    @staticmethod
    def parse_datetime(time):
        """
        Parses and verifies tile sensing time.

        :param time: tile sensing time
        :type time: str
        :return: tile sensing time in ISO8601 format
        :rtype: str
        """
        try:
            return parse_time(time)
        except Exception:
            raise ValueError('Time must be in format YYYY-MM-DD or YYYY-MM-DDTHH:MM:SS')

    def get_requests(self):
        """
        Creates tile structure and returns list of files for download.

        :return: List of download requests and list of empty folders that need to be created
        :rtype: (list(download.DownloadRequest), list(str))
        """
        self.download_list = []
        for data_name in [band for band in self.bands if self._band_exists(band)] + self.metafiles:
            if data_name in AwsConstants.TILE_FILES:
                url = self.get_url(data_name)
                filename = self.get_filepath(data_name)
                self.download_list.append(DownloadRequest(url=url, filename=filename,
                                                          data_type=AwsConstants.AWS_FILES[data_name],
                                                          data_name=data_name))
        self.sort_download_list()
        return self.download_list, self.folder_list

    def get_aws_index(self):
        """
        Returns tile index on AWS. If `tile_index` was not set during class initialization it will be determined
        according to existing tiles on AWS.

        :return: Index of tile on AWS
        :rtype: int
        """
        if self.aws_index is not None:
            return self.aws_index
        tile_info_list = get_tile_info(self.tile_name, self.datetime, all_tiles=True)
        if not tile_info_list:
            raise ValueError('Cannot find aws_index for specified tile and time')

        if self.data_source is DataSource.SENTINEL2_L2A:
            for tile_info in sorted(tile_info_list, key=self._parse_aws_index):
                try:
                    self.aws_index = self._parse_aws_index(tile_info)
                    self.get_tile_info()
                    return self.aws_index
                except AwsDownloadFailedException:
                    pass

        return self._parse_aws_index(tile_info_list[0])

    @staticmethod
    def _parse_aws_index(tile_info):
        """Parses AWS index from tile info

        :param tile_info: dictionary with information about tile
        :type tile_info: dict
        :return: Index of tile on AWS
        :rtype: int
        """
        return int(tile_info['properties']['s3Path'].split('/')[-1])

    def tile_is_valid(self):
        """ Checks if tile has tile info and valid timestamp

        :return: `True` if tile is valid and `False` otherwise
        :rtype: bool
        """
        return self.tile_info is not None \
            and (self.datetime == self.date or self.datetime == self.parse_datetime(self.tile_info['timestamp']))

    def get_tile_info(self):
        """
        Collects basic info about tile from tileInfo.json.

        :return: dictionary with tile information
        :rtype: dict
        """
        return get_aws_json(self.get_url(AwsConstants.TILE_INFO))

    def get_url(self, filename):
        """
        Creates url of file location on AWS.

        :param filename: name of file
        :type filename: str
        :return: url of file location
        :rtype: str
        """
        tile_url = self.tile_url
        force_http = filename in [AwsConstants.TILE_INFO, AwsConstants.PRODUCT_INFO, AwsConstants.METADATA]
        if tile_url is None or force_http:
            tile_url = self.get_tile_url(force_http=force_http)
        return '{}/{}'.format(tile_url, self.add_file_extension(filename))

    def get_tile_url(self, force_http=False):
        """
        Creates base url of tile location on AWS.

        :param force_http: `True` if HTTP base URL should be used and `False` otherwise
        :type force_http: str
        :return: url of tile location
        :rtype: str
        """
        base_url = self.base_http_url if force_http else self.base_url
        url = '{}/tiles/{}/{}/{}/'.format(base_url, self.tile_name[0:2].lstrip('0'), self.tile_name[2],
                                          self.tile_name[3:5])
        date_params = self.date.split('-')
        for param in date_params:
            url += param.lstrip('0') + '/'
        return url + str(self.aws_index)

    def get_qi_url(self, metafile):
        """Returns url of tile metadata products

        :param metafile: Name of metadata product at AWS
        :type metafile: str
        :return: url location of metadata product at AWS
        :rtype: str
        """
        return '{}/qi/{}'.format(self.tile_url, metafile)

    def get_gml_url(self, qi_type, band='B00'):
        """
        :param qi_type: type of quality indicator
        :type qi_type: str
        :param band: band name
        :type band: str
        :return: location of gml file on AWS
        :rtype: str
        """
        band = band.split('/')[-1]
        return self.get_qi_url('MSK_{}_{}.gml'.format(qi_type, band))

    def get_preview_url(self, data_type='L1C'):
        """Returns url location of full resolution L1C preview
        :return:
        """
        if self.data_source is DataSource.SENTINEL2_L1C or self.safe_type is EsaSafeType.OLD_TYPE:
            return self.get_url(AwsConstants.PREVIEW_JP2)
        return self.get_qi_url('{}_PVI.jp2'.format(data_type))

    def get_filepath(self, filename):
        """
        Creates file path for the file.

        :param filename: name of the file
        :type filename: str
        :return: filename with path on disk
        :rtype: str
        """
        return os.path.join(self.parent_folder, '{},{},{}'.format(self.tile_name, self.date, self.aws_index),
                            self.add_file_extension(filename)).replace(':', '.')

    def get_product_id(self):
        """
        Obtains ESA ID of product which contains the tile.

        :return: ESA ID of the product
        :rtype: str
        """
        return self.tile_info['productName']

    def _band_exists(self, band_name):
        if self.data_source is DataSource.SENTINEL2_L1C:
            return True
        resolution, band = band_name.split('/')
        if self.safe_type is EsaSafeType.COMPACT_TYPE:
            return not (self.baseline >= '02.07' and band.endswith(AwsConstants.VIS))
        return band != AwsConstants.TCI and not (band == AwsConstants.SCL and resolution == AwsConstants.R60m)

    @staticmethod
    def tile_id_to_tile(tile_id):
        """
        :param tile_id: original tile identification string provided by ESA (e.g.
                        'S2A_OPER_MSI_L1C_TL_SGS__20160109T230542_A002870_T10UEV_N02.01')
        :type tile_id: str
        :return: tile name, sensing date and AWS index
        :rtype: (str, str, int)
        """
        if tile_id.split('_')[0] not in ['S2A', 'S2B', 'L1C']:
            raise ValueError('Transformation from tile ID to tile works currently only for Sentinel-2 L1C products')

        tile_info = get_tile_info_id(tile_id)
        return AwsService.url_to_tile(tile_info['properties']['s3Path'])



"""
Module implementing rate limiting logic for Sentinel Hub service
"""
import math
import time
from enum import Enum

from .download.client import get_json
from .exceptions import OutOfRequestsException


ERR = 0.01

class SentinelHubRateLimit:
    """ Class implementing rate limiting logic of Sentinel Hub service

    It has 2 public methods:
     - register_next - tells if next download can start or if not, what is the wait before it can be asked again
     - update - updates expectations according to headers obtained from download

    The rate limiting object is collecting information about the status of rate limiting policy buckets from
    Sentinel Hub service. According to this information and a feedback from download requests it adapts expectations
    about when the next download attempt will be possible.
    """

    REQUEST_RETRY_HEADER = 'Retry-After'
    REQUEST_COUNT_HEADER = 'X-RateLimit-Remaining'
    UNITS_RETRY_HEADER = 'X-ProcessingUnits-Retry-After'
    UNITS_COUNT_HEADER = 'X-ProcessingUnits-Remaining'
    VIOLATION_HEADER = 'X-RateLimit-ViolatedPolicy'

    def __init__(self, session, *, status_refresh_period=60):
        """
        :param session: An instance of Sentinel Hub session object
        :type session: sentinelhub.SentinelHubSession or None
        :param status_refresh_period: Number of seconds between two consecutive rate limit status checks
        :type status_refresh_period: int
        """
        self.session = session
        self.status_refresh_period = None if self.session is None else status_refresh_period

        self.requests_completed = 0  # This counts completed requests only in a single status check period
        self.requests_in_process = 0

        # Initial expectations:
        self.expected_process_num = 1
        self.expected_cost_per_request = 1

        self.last_status_update_time = time.monotonic() if self.session is None else None
        self.policy_buckets = self._initialize_policy_buckets()

    def _initialize_policy_buckets(self):
        """ Collects and prepares a list of policy buckets defined for a user for which a given session has been
        created
        """
        if self.session is None:
            return self._initialize_trial_policy_buckets()

        user_policies = self._fetch_user_policies()

        if not user_policies['data']:
            user_policies = self._fetch_user_policies(default_policies=True)

        buckets = []
        for policy_payload in user_policies['data']:

            policy_info = policy_payload['type'] if 'type' in policy_payload else policy_payload
            policy_type = policy_info['name']

            policies = policy_payload.get('policies')  # Can either be None or an empty list
            if not policies:
                policies = policy_info['defaultPolicies']

            for policy in policies:
                buckets.append(PolicyBucket(policy_type, policy))

        return buckets

    @staticmethod
    def _initialize_trial_policy_buckets():
        """ Prepares default trial policy buckets
        """
        return [
            PolicyBucket(PolicyType.PROCESSING_UNITS, {
                "capacity": 30000,
                "samplingPeriod": "PT744H",
                "nanosBetweenRefills": 89280000000
            }),
            PolicyBucket(PolicyType.PROCESSING_UNITS, {
                "capacity": 300,
                "samplingPeriod": "PT1M",
                "nanosBetweenRefills": 200000000,
            }),
            PolicyBucket(PolicyType.REQUESTS, {
                "capacity": 30000,
                "samplingPeriod": "PT744H",
                "nanosBetweenRefills": 89280000000,
            }),
            PolicyBucket(PolicyType.REQUESTS, {
                "capacity": 300,
                "samplingPeriod": "PT1M",
                "nanosBetweenRefills": 200000000,
            })
        ]

    def _fetch_user_policies(self, default_policies=False):
        """ Collects user rate limiting policies
        """
        url = '{}/contract'.format(self.session.config.get_sh_rate_limit_url())
        if default_policies:
            url = '{}/types'.format(url)

        return get_json(url, headers=self.session.session_headers)

    def _fetch_status(self):
        """ Collects status about remaining requests and processing units
        """
        url = '{}/statistics/tokenCounts'.format(self.session.config.get_sh_rate_limit_url())

        return get_json(url, headers=self.session.session_headers)

    def register_next(self):
        """ Determines if next download request can start or not
        """
        if self.status_refresh_period is not None and \
                (self.last_status_update_time is None or
                 self.last_status_update_time + self.status_refresh_period <= time.monotonic()):
            self._refresh_status()

        elapsed_time = time.monotonic() - self.last_status_update_time

        max_wait_time = 0
        for bucket in self.policy_buckets:
            expected_cost_per_request = 1 if bucket.is_request_bucket() else self.expected_cost_per_request
            max_requests_completed = self.requests_completed + self.requests_in_process

            expected_wait_time = bucket.get_expected_wait_time(elapsed_time,
                                                               self.expected_process_num,
                                                               expected_cost_per_request,
                                                               max_requests_completed)
            if expected_wait_time == -1:
                raise OutOfRequestsException('Your user has run out of available number of requests')

            max_wait_time = max(max_wait_time, expected_wait_time)

        if max_wait_time == 0:
            self.requests_in_process += 1
        return max_wait_time

    def _refresh_status(self):
        """ Collects new status, updates expectations and updates buckets with the new content
        """
        new_status = self._fetch_status()

        new_status_update_time = time.monotonic()
        new_content_list = [new_status['data'][bucket.policy_type.value][bucket.refill_period]
                            for bucket in self.policy_buckets]

        if self.last_status_update_time is not None:
            elapsed_time = new_status_update_time - self.last_status_update_time
            self._update_expectations(elapsed_time, new_content_list)

        for bucket, new_content in zip(self.policy_buckets, new_content_list):
            bucket.content = new_content

        self.last_status_update_time = new_status_update_time
        self.requests_completed = 0

    def _update_expectations(self, elapsed_time, new_content_list):
        """ Updates expectation variables
        """

        cost_per_second_list = [bucket.count_cost_per_second(elapsed_time, new_content) for bucket, new_content in
                                zip(self.policy_buckets, new_content_list)]

        requests_per_second = min(cost for cost, bucket in zip(cost_per_second_list, self.policy_buckets)
                                  if bucket.is_request_bucket())
        cost_per_second = min(cost for cost, bucket in zip(cost_per_second_list, self.policy_buckets)
                              if not bucket.is_request_bucket())

        cost_per_request = cost_per_second / max(requests_per_second, ERR)

        self._update_expected_process_num(elapsed_time * requests_per_second)
        self._update_expected_cost_per_request(cost_per_request)

    def update(self, headers):
        """ Update expectations by using information from response headers
        """
        elapsed_time = time.monotonic() - self.last_status_update_time
        is_rate_limited = self.VIOLATION_HEADER in headers

        self.requests_in_process -= 1
        if not is_rate_limited:
            self.requests_completed += 1

        if self.REQUEST_COUNT_HEADER not in headers:
            return

        remaining_requests = float(headers[self.REQUEST_COUNT_HEADER])
        requests_per_second = max(bucket.count_cost_per_second(elapsed_time, remaining_requests) for bucket in
                                  self.policy_buckets if bucket.is_request_bucket())

        self._update_expected_process_num(elapsed_time * requests_per_second, is_rate_limited=is_rate_limited)

        if self.UNITS_COUNT_HEADER not in headers:
            return

        remaining_units = float(headers[self.UNITS_COUNT_HEADER])
        cost_per_second = max(bucket.count_cost_per_second(elapsed_time, remaining_units) for bucket in
                              self.policy_buckets if not bucket.is_request_bucket())
        cost_per_request = cost_per_second / max(requests_per_second, ERR)

        self._update_expected_cost_per_request(cost_per_request, is_rate_limited=is_rate_limited)

    def _update_expected_process_num(self, requests_done, is_rate_limited=False):
        """ This method updates the expected number of concurrent processes using this credentials
        """
        # Note: if self.requests_completed is small the following is likely to be inaccurate
        max_process_num = math.ceil(requests_done / max(self.requests_completed + self.requests_in_process, 1))
        min_process_num = math.floor(requests_done / max(self.requests_completed + self.requests_in_process, 1))

        if is_rate_limited or self.expected_process_num < min_process_num:
            self.expected_process_num += 1

        if self.expected_process_num > max_process_num:
            self.expected_process_num -= 1

        self.expected_process_num = max(self.expected_process_num, 1)

    def _update_expected_cost_per_request(self, cost_per_request, is_rate_limited=False):
        """ This method updates the expected cost per request
        """
        value_difference = cost_per_request - self.expected_cost_per_request

        if is_rate_limited:
            # Making sure we cannot decrease expectations if we get rate-limited
            value_difference = max(value_difference, 1)
        else:
            value_difference = min(max(value_difference, -1), 1) / 2

        self.expected_cost_per_request += value_difference
        self.expected_cost_per_request = max(self.expected_cost_per_request, 1)


class PolicyBucket:
    """ A class representing Sentinel Hub policy bucket
    """
    def __init__(self, policy_type, policy_payload):
        """
        :param policy_type: A type of policy
        :type policy_type: PolicyType or str
        :param policy_payload: A dictionary of policy parameters
        :type policy_payload: dict
        """

        self.policy_type = PolicyType(policy_type)

        self.capacity = float(policy_payload['capacity'])
        self.refill_period = policy_payload['samplingPeriod']

        # The following is the same as if we would interpret samplingPeriod string
        self.refill_per_second = 10 ** 9 / policy_payload['nanosBetweenRefills']

        self._content = self.capacity

    def __repr__(self):
        """ Representation of the bucket content
        """
        return '{}(policy_type={}, content={}/{}, refill_period={}, refill_per_second={})' \
               ''.format(self.__class__.__name__, self.policy_type, self.content, self.capacity, self.refill_period,
                         self.refill_per_second)

    @property
    def content(self):
        """ Variable `content` can be accessed as a property
        """
        return self._content

    @content.setter
    def content(self, value):
        """ Variable `content` can be modified by external classes
        """
        self._content = value

    def count_cost_per_second(self, elapsed_time, new_content):
        """ Calculates the cost per second for the bucket given the elapsed time and the new content.

        In the calculation it assumes that during the elapsed time bucket was being filled all the time - i.e. it
        assumes the bucket has never been ful for a non-zero amount of time in the elapsed time period.
        """
        content_difference = self.content - new_content
        if not self.is_fixed():
            content_difference += elapsed_time * self.refill_per_second

        return content_difference / elapsed_time

    def get_expected_wait_time(self, elapsed_time, expected_process_num, expected_cost_per_request,
                               requests_completed, buffer_cost=0.5):
        """ Expected time a user would have to wait for this bucket
        """
        overall_completed_cost = requests_completed * expected_cost_per_request * expected_process_num
        expected_content = self.content + elapsed_time * self.refill_per_second - overall_completed_cost

        if self.is_fixed():
            if expected_content < expected_cost_per_request:
                return -1
            return 0

        return max(expected_cost_per_request - expected_content + buffer_cost, 0) / self.refill_per_second

    def is_request_bucket(self):
        """ Checks if bucket counts requests
        """
        return self.policy_type is PolicyType.REQUESTS

    def is_fixed(self):
        """ Checks if bucket has a fixed number of requests
        """
        return self.refill_period == 'PT0S'


class PolicyType(Enum):
    """ Enum defining different types of policies
    """
    PROCESSING_UNITS = 'PROCESSING_UNITS'
    REQUESTS = 'REQUESTS'



"""
Utility functions to read/write image data from/to file
"""

import csv
import json
import os
import logging
import warnings
from xml.etree import ElementTree

import numpy as np
import tifffile as tiff
from PIL import Image

from .decoding import decode_tar, get_data_format, fix_jp2_image, get_jp2_bit_depth
from .constants import MimeType
from .os_utils import create_parent_folder


warnings.simplefilter('ignore', Image.DecompressionBombWarning)
LOGGER = logging.getLogger(__name__)

CSV_DELIMITER = ';'


class LocalIO:
    """ The default IO class for reading and writing the data
    """
    @staticmethod
    def read(filename, **kwargs):
        """ Function implementing the reading logic
        """
        return read_data(filename, **kwargs)

    @staticmethod
    def write(filename, data, **kwargs):
        """ Function implementing the writing logic
        """
        return write_data(filename, data, **kwargs)

    @staticmethod
    def exists(file_path):
        """ Function that checks if a given file exists
        """
        return os.path.exists(file_path)


def read_data(filename, data_format=None):
    """ Read image data from file

    This function reads input data from file. The format of the file
    can be specified in ``data_format``. If not specified, the format is
    guessed from the extension of the filename.

    :param filename: filename to read data from
    :type filename: str
    :param data_format: format of filename. Default is `None`
    :type data_format: MimeType
    :return: data read from filename
    :raises: exception if filename does not exist
    """
    if not os.path.exists(filename):
        raise ValueError('Filename {} does not exist'.format(filename))

    if not isinstance(data_format, MimeType):
        data_format = get_data_format(filename)

    if data_format.is_tiff_format():
        return read_tiff_image(filename)
    if data_format is MimeType.JP2:
        return read_jp2_image(filename)
    if data_format.is_image_format():
        return read_image(filename)
    try:
        return {
            MimeType.TAR: read_tar,
            MimeType.TXT: read_text,
            MimeType.CSV: read_csv,
            MimeType.JSON: read_json,
            MimeType.XML: read_xml,
            MimeType.GML: read_xml,
            MimeType.SAFE: read_xml
        }[data_format](filename)
    except KeyError:
        raise ValueError('Reading data format .{} is not supported'.format(data_format.value))

def read_tar(filename):
    """ Read a tar from file
    """
    with open(filename, 'rb') as file:
        return decode_tar(file.read())

def read_tiff_image(filename):
    """ Read data from TIFF file

    :param filename: name of TIFF file to be read
    :type filename: str
    :return: data stored in TIFF file
    """
    return tiff.imread(filename)


def read_jp2_image(filename):
    """ Read data from JPEG2000 file

    :param filename: name of JPEG2000 file to be read
    :type filename: str
    :return: data stored in JPEG2000 file
    """
    # Other option:
    # return glymur.Jp2k(filename)[:]
    image = read_image(filename)

    with open(filename, 'rb') as file:
        bit_depth = get_jp2_bit_depth(file)

    return fix_jp2_image(image, bit_depth)


def read_image(filename):
    """ Read data from PNG or JPG file

    :param filename: name of PNG or JPG file to be read
    :type filename: str
    :return: data stored in JPG file
    """
    return np.array(Image.open(filename))


def read_text(filename):
    """ Read data from text file

    :param filename: name of text file to be read
    :type filename: str
    :return: data stored in text file
    """
    with open(filename, 'r') as file:
        return file.read()   # file.readline() for reading 1 line


def read_csv(filename, delimiter=CSV_DELIMITER):
    """ Read data from CSV file

    :param filename: name of CSV file to be read
    :type filename: str
    :param delimiter: type of CSV delimiter. Default is ``;``
    :type delimiter: str
    :return: data stored in CSV file as list
    """
    with open(filename, 'r') as file:
        return list(csv.reader(file, delimiter=delimiter))


def read_json(filename):
    """ Read data from JSON file

    :param filename: name of JSON file to be read
    :type filename: str
    :return: data stored in JSON file
    """
    with open(filename, 'r') as file:
        return json.load(file)


def read_xml(filename):
    """ Read data from XML or GML file

    :param filename: name of XML or GML file to be read
    :type filename: str
    :return: data stored in XML file
    """
    return ElementTree.parse(filename)


def read_numpy(filename):
    """ Read data from numpy file

    :param filename: name of numpy file to be read
    :type filename: str
    :return: data stored in file as numpy array
    """
    return np.load(filename)


def write_data(filename, data, data_format=None, compress=False, add=False):
    """ Write image data to file

    Function to write image data to specified file. If file format is not provided
    explicitly, it is guessed from the filename extension. If format is TIFF, geo
    information and compression can be optionally added.

    :param filename: name of file to write data to
    :type filename: str
    :param data: image data to write to file
    :type data: numpy array
    :param data_format: format of output file. Default is `None`
    :type data_format: MimeType
    :param compress: whether to compress data or not. Default is `False`
    :type compress: bool
    :param add: whether to append to existing text file or not. Default is `False`
    :type add: bool
    :raises: exception if numpy format is not supported or file cannot be written
    """
    create_parent_folder(filename)

    if not isinstance(data_format, MimeType):
        data_format = get_data_format(filename)

    if data_format.is_tiff_format():
        return write_tiff_image(filename, data, compress)
    if data_format.is_image_format():
        return write_image(filename, data)
    if data_format is MimeType.TXT:
        return write_text(filename, data, add=add)

    try:
        return {
            MimeType.RAW: write_bytes,
            MimeType.CSV: write_csv,
            MimeType.JSON: write_json,
            MimeType.XML: write_xml,
            MimeType.GML: write_xml
        }[data_format](filename, data)
    except KeyError:
        raise ValueError('Writing data format .{} is not supported'.format(data_format.value))


def write_tiff_image(filename, image, compress=False):
    """ Write image data to TIFF file

    :param filename: name of file to write data to
    :type filename: str
    :param image: image data to write to file
    :type image: numpy array
    :param compress: whether to compress data. If `True`, lzma compression is used. Default is `False`
    :type compress: bool
    """
    if compress:
        return tiff.imsave(filename, image, compress='lzma')  # loseless compression, works very well on masks
    return tiff.imsave(filename, image)


def write_jp2_image(filename, image):
    """ Write image data to JPEG2000 file

    :param filename: name of JPEG2000 file to write data to
    :type filename: str
    :param image: image data to write to file
    :type image: numpy array
    :return: jp2k object
    """
    # Other options:
    # return glymur.Jp2k(filename, data=image)
    # cv2.imwrite(filename, image)
    return write_image(filename, image)


def write_image(filename, image):
    """ Write image data to PNG, JPG file

    :param filename: name of PNG or JPG file to write data to
    :type filename: str
    :param image: image data to write to file
    :type image: numpy array
    """
    data_format = get_data_format(filename)
    if data_format is MimeType.JPG:
        LOGGER.warning('Warning: jpeg is a lossy format therefore saved data will be modified.')
    return Image.fromarray(image).save(filename)


def write_text(filename, data, add=False):
    """ Write image data to text file

    :param filename: name of text file to write data to
    :type filename: str
    :param data: image data to write to text file
    :type data: numpy array
    :param add: whether to append to existing file or not. Default is `False`
    :type add: bool
    """
    write_type = 'a' if add else 'w'
    with open(filename, write_type) as file:
        print(data, end='', file=file)


def write_csv(filename, data, delimiter=CSV_DELIMITER):
    """ Write image data to CSV file

    :param filename: name of CSV file to write data to
    :type filename: str
    :param data: image data to write to CSV file
    :type data: numpy array
    :param delimiter: delimiter used in CSV file. Default is ``;``
    :type delimiter: str
    """
    with open(filename, 'w') as file:
        csv_writer = csv.writer(file, delimiter=delimiter)
        for line in data:
            csv_writer.writerow(line)


def write_json(filename, data):
    """ Write data to JSON file

    :param filename: name of JSON file to write data to
    :type filename: str
    :param data: data to write to JSON file
    :type data: list, tuple
    """
    with open(filename, 'w') as file:
        json.dump(data, file, indent=4, sort_keys=True)


def write_xml(filename, element_tree):
    """ Write data to XML or GML file

    :param filename: name of XML or GML file to write data to
    :type filename: str
    :param element_tree: data as ElementTree object
    :type element_tree: xmlElementTree
    """
    return element_tree.write(filename)
    # this will write declaration tag in first line:
    # return element_tree.write(filename, encoding='utf-8', xml_declaration=True)


def write_numpy(filename, data):
    """ Write data as numpy file

    :param filename: name of numpy file to write data to
    :type filename: str
    :param data: data to write to numpy file
    :type data: numpy array
    """
    return np.save(filename, data)


def write_bytes(filename, data):
    """ Write binary data into a file

    :param filename:
    :param data:
    :return:
    """
    with open(filename, 'wb') as file:
        file.write(data)



import io
import os
from setuptools import setup, find_packages


def parse_requirements(file):
    required_packages = []
    with open(os.path.join(os.path.dirname(__file__), file)) as req_file:
        for line in req_file:
            if '/' not in line:
                required_packages.append(line.strip())
    return required_packages


def get_version():
    for line in open(os.path.join(os.path.dirname(__file__), 'sentinelhub', '_version.py')):
        if line.find("__version__") >= 0:
            version = line.split("=")[1].strip()
            return version.strip('"').strip("'")


def get_long_description():
    return io.open('README.md', encoding="utf-8").read()


def update_package_config():
    """ Every time sentinelhub package is installed entire config.json is overwritten. However this function
    will check if sentinelhub is already installed and try to copy those parameters from old config.json that are by
    default set to an empty value (i.e. instance_id, aws_access_key_id and aws_secret_access_key) into new config.json
    file.
    """
    try:
        import importlib
        import sys
        import json

        path = importlib.machinery.PathFinder().find_spec('sentinelhub', sys.path[1:]).submodule_search_locations[0]
        old_config_filename = os.path.join(path, 'config.json')

        with open(old_config_filename, 'r') as file:
            old_config = json.load(file)

        from sentinelhub.config import SHConfig

        config = SHConfig()
        for attr, value in old_config.items():
            if hasattr(config, attr) and not getattr(config, attr):
                setattr(config, attr, value)

        config.save()

    except BaseException:
        pass


update_package_config()

setup(
    name='sentinelhub',
    python_requires='>=3.5',
    version=get_version(),
    description='Sentinel Hub Utilities',
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url='https://github.com/sentinel-hub/sentinelhub-py',
    author='Sinergise ltd.',
    author_email='info@sentinel-hub.com',
    license='MIT',
    packages=find_packages(),
    package_data={'sentinelhub': ['sentinelhub/config.json', 'sentinelhub/.utmzones.geojson']},
    include_package_data=True,
    install_requires=parse_requirements('requirements.txt'),
    extras_require={
        'DEV': parse_requirements('requirements-dev.txt'),
        'DOCS': parse_requirements('requirements-docs.txt')
    },
    zip_safe=False,
    entry_points={'console_scripts': ['sentinelhub=sentinelhub.commands:main_help',
                                      'sentinelhub.aws=sentinelhub.commands:aws',
                                      'sentinelhub.config=sentinelhub.commands:config',
                                      'sentinelhub.download=sentinelhub.commands:download']},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development'
    ]
)



"""
Module for working with Geopedia
"""

import logging
import datetime
import hashlib

from shapely.geometry import shape as geo_shape

from .ogc import OgcImageService, MimeType
from .config import SHConfig
from .download import DownloadRequest, get_json
from .constants import CRS

LOGGER = logging.getLogger(__name__)


class GeopediaService:
    """ The class for Geopedia OGC services
    """
    def __init__(self, base_url=None):
        """
        :param base_url: Base url of Geopedia REST services. If `None`, the url specified in the configuration
                    file is taken.
        :type base_url: str or None
        """
        self.base_url = SHConfig().geopedia_rest_url if base_url is None else base_url

    @staticmethod
    def _parse_layer(layer, return_wms_name=False):
        """ Helper function for parsing Geopedia layer name. If WMS name is required and wrong form is given it will
        return a string with 'ttl' at the beginning. (WMS name can also start with something else, e.g. only 't'
        instead 'ttl', therefore anything else is also allowed.) Otherwise it will parse it into a number.
        """
        if not isinstance(layer, (int, str)):
            raise ValueError("Parameter 'layer' should be an integer or a string, but {} found".format(type(layer)))

        if return_wms_name:
            if isinstance(layer, int) or layer.isdigit():
                return 'ttl{}'.format(layer)
            return layer

        if isinstance(layer, str):
            stripped_layer = layer.lstrip('tl')
            if not stripped_layer.isdigit():
                raise ValueError("Parameter 'layer' has unsupported value {}, expected an integer".format(layer))
            layer = stripped_layer

        return int(layer)


class GeopediaSession(GeopediaService):
    """ For retrieving data from Geopedia vector and raster layers it is required to make a session. This class handles
    starting and renewing of session and login (optional). It provides session headers required by Geopedia REST
    requests. Session duration is hardcoded to 1 hour with class attribute SESSION_DURATION. After that this
    class will automatically renew the session and login.
    """
    SESSION_DURATION = datetime.timedelta(hours=1)
    UNAUTHENTICATED_USER_ID = 'NO_USER'

    _global_session_info = None
    _global_session_start = None

    def __init__(self, *, username=None, password=None, password_md5=None, is_global=False, **kwargs):
        """
        :param username: Optional parameter in case of login with Geopedia credentials
        :type username: str or None
        :param password: Optional parameter in case of login with Geopedia credentials
        :type password: str or None
        :param password_md5: Password can optionally also be provided as already encoded md5 hexadecimal string
        :type password_md5: str or None
        :param is_global: If `True` this session will be shared among all instances of this class, otherwise it will be
            used only with the single instance. Default is `False`.
        :type is_global: bool
        :param base_url: Base url of Geopedia REST services. If `None`, the url specified in the configuration
            file is taken.
        :type base_url: str or None
        """
        super().__init__(**kwargs)

        if password and password_md5:
            raise ValueError("At most one of the parameters 'password' and 'password_md5' can be specified, not both")

        self.username = username
        self.password = password_md5 if password is None else hashlib.md5(password.encode()).hexdigest()
        self.is_global = is_global

        if bool(self.username) != bool(self.password):
            raise ValueError('Either both username and password have to be specified or neither of them, '
                             'only one found')

        self._session_info = None
        self._session_start = None

        self.provide_session()

    @property
    def session_info(self):
        """ All information that Geopedia provides about the current session

        :return: A dictionary with session info
        :rtype: dict
        """
        return self.provide_session()

    @property
    def session_id(self):
        """ A public property of this class which provides a Geopedia session ID

        :return: A session ID string
        :rtype: str
        """
        return self._parse_session_id(self.provide_session())

    @property
    def session_headers(self):
        """ Headers which have to be used when accessing any data from Geopedia with this session

        :return: A dictionary containing session headers
        :rtype: dict
        """
        session_info = self.provide_session()
        return {
            session_info['sessionHeaderName']: self._parse_session_id(session_info)
        }

    @property
    def user_info(self):
        """ Information that this session has about user

        :return: A dictionary with user info
        :rtype: dict
        """
        return self.provide_session()['user']

    @property
    def user_id(self):
        """ Geopedia user ID. If no login was done during session this will return `'NO_USER'`.

        :return: User ID string
        :rtype: str
        """
        return self._parse_user_id(self.provide_session())

    def restart(self):
        """ Method that restarts Geopedia Session

        :return: It returns the object itself, with new session
        :rtype: GeopediaSession
        """
        self.provide_session(start_new=True)
        return self

    def provide_session(self, start_new=False):
        """ Makes sure that session is still valid and provides session info

        :param start_new: If `True` it will always create a new session. Otherwise it will create a new
            session only if no session exists or the previous session timed out.
        :type start_new: bool
        :return: Current session info
        :rtype: dict
        """
        if self.is_global:
            self._session_info = self._global_session_info
            self._session_start = self._global_session_start

        if self._session_info is None or start_new or \
                datetime.datetime.now() > self._session_start + self.SESSION_DURATION:
            self._start_new_session()

        return self._session_info

    def _start_new_session(self):
        """ Starts a new session and calculates when the new session will end. If username and password are provided
        it will also make login.
        """
        self._session_start = datetime.datetime.now()

        session_id = self._parse_session_id(self._session_info) if self._session_info else ''
        session_url = '{}/data/v1/session/create?locale=en&sid={}'.format(self.base_url, session_id)
        self._session_info = get_json(session_url)

        if self.username and self.password and self._parse_user_id(self._session_info) == self.UNAUTHENTICATED_USER_ID:
            self._make_login()

        if self.is_global:
            GeopediaSession._global_session_info = self._session_info
            GeopediaSession._global_session_start = self._session_start

    def _make_login(self):
        """ Private method that makes login
        """
        login_url = '{}/data/v1/session/login?user={}&pass={}&sid={}'.format(self.base_url, self.username,
                                                                             self.password,
                                                                             self._parse_session_id(self._session_info))
        self._session_info = get_json(login_url)

    @staticmethod
    def _parse_session_id(session_info):
        """ Method for parsing session ID from session info
        """
        return session_info['sessionId']

    @staticmethod
    def _parse_user_id(session_info):
        """ Method for parsing user ID from session info
        """
        return session_info['user']['id']


class GeopediaWmsService(GeopediaService, OgcImageService):
    """ Geopedia OGC services class for providing image data. Most of the methods are inherited from
    `sentinelhub.ogc.OgcImageService` class.
    """
    def __init__(self, base_url=None):
        """
        :param base_url: Base url of Geopedia WMS services. If `None`, the url specified in the configuration
                    file is taken.
        :type base_url: str or None
        """
        super().__init__(base_url=SHConfig().geopedia_wms_url if base_url is None else base_url)

    def get_request(self, request):
        """ Get a list of DownloadRequests for all data that are under the given field in the table of a Geopedia layer.

        :return: list of items which have to be downloaded
        :rtype: list(DownloadRequest)
        """
        request.layer = self._parse_layer(request.layer, return_wms_name=True)

        return super().get_request(request)

    def get_dates(self, request):
        """ Geopedia does not support date queries

        :param request: OGC-type request
        :type request: WmsRequest or WcsRequest
        :return: Undefined date
        :rtype: [None]
        """
        return [None]

    def get_wfs_iterator(self):
        """ This method is inherited from OgcImageService but is not implemented.
        """
        raise NotImplementedError


class GeopediaImageService(GeopediaService):
    """ Service class that provides images from a Geopedia vector layer.
    """
    def __init__(self, **kwargs):
        """
        :param base_url: Base url of Geopedia REST services. If `None`, the url
                     specified in the configuration file is taken.
        :type base_url: str or None
        """
        super().__init__(**kwargs)

        self.gpd_iterator = None

    def get_request(self, request):
        """ Get a list of DownloadRequests for all data that are under the given field in the table of a Geopedia layer.

        :return: list of items which have to be downloaded
        :rtype: list(DownloadRequest)
        """
        return [DownloadRequest(url=self._get_url(item),
                                filename=self._get_filename(request, item),
                                data_type=request.image_format)
                for item in self._get_items(request)]

    def _get_items(self, request):
        """ Collects data from Geopedia layer and returns list of features
        """
        if request.gpd_iterator is None:
            self.gpd_iterator = GeopediaFeatureIterator(request.layer, bbox=request.bbox, base_url=self.base_url,
                                                        gpd_session=request.gpd_session)
        else:
            self.gpd_iterator = request.gpd_iterator

        field_iter = self.gpd_iterator.get_field_iterator(request.image_field_name)
        items = []

        for field_items in field_iter:  # an image field can have multiple images

            for item in field_items:
                if not item['mimeType'].startswith('image/'):
                    continue

                mime_type = MimeType.from_string(item['mimeType'][6:])

                if mime_type is request.image_format:
                    items.append(item)

        return items

    @staticmethod
    def _get_url(item):
        return item.get('objectPath')

    @staticmethod
    def _get_filename(request, item):
        """ Creates a filename
        """
        if request.keep_image_names:
            filename = OgcImageService.finalize_filename(item['niceName'].replace(' ', '_'))
        else:
            filename = OgcImageService.finalize_filename(
                '_'.join([str(GeopediaService._parse_layer(request.layer)), item['objectPath'].rsplit('/', 1)[-1]]),
                request.image_format
            )

        LOGGER.debug("filename=%s", filename)
        return filename

    def get_gpd_iterator(self):
        """ Returns iterator over info about data used for the `GeopediaVectorRequest`

        :return: Iterator of dictionaries containing info about data used in the request.
        :rtype: Iterator[dict] or None
        """
        return self.gpd_iterator


class GeopediaFeatureIterator(GeopediaService):
    """ Iterator for Geopedia Vector Service
    """
    FILTER_EXPRESSION = 'filterExpression'

    def __init__(self, layer, bbox=None, query_filter=None, gpd_session=None, **kwargs):
        """
        :param layer: Geopedia layer which contains requested data
        :type layer: str
        :param bbox: Bounding box of the requested image. Its coordinates must be in the CRS.POP_WEB (EPSG:3857)
            coordinate system.
        :type bbox: BBox
        :param query_filter: Query string used for filtering returned features.
        :type query_filter: str
        :param gpd_session: Optional parameter for specifying a custom Geopedia session, which can also contain login
            credentials. This can be used for accessing private Geopedia layers. By default it is set to `None` and a
            basic Geopedia session without credentials will be created.
        :type gpd_session: GeopediaSession or None
        :param base_url: Base url of Geopedia REST services. If `None`, the url specified in the configuration
            file is taken.
        :type base_url: str or None
        """
        super().__init__(**kwargs)

        self.layer = self._parse_layer(layer)

        self.query = {}
        if bbox is not None:
            if bbox.crs is not CRS.POP_WEB:
                bbox = bbox.transform(CRS.POP_WEB)

            self.query[self.FILTER_EXPRESSION] = 'bbox({},"EPSG:3857")'.format(bbox)
        if query_filter is not None:
            if self.FILTER_EXPRESSION in self.query:
                self.query[self.FILTER_EXPRESSION] = '{} && ({})'.format(self.query[self.FILTER_EXPRESSION],
                                                                         query_filter)
            else:
                self.query[self.FILTER_EXPRESSION] = query_filter

        self.gpd_session = gpd_session if gpd_session else GeopediaSession(is_global=True)
        self.features = []
        self.layer_size = None
        self.index = 0

        self.next_page_url = '{}/data/v2/search/tables/{}/features'.format(self.base_url, self.layer)

    def __iter__(self):
        self.index = 0

        return self

    def __next__(self):
        if self.index == len(self.features):
            self._fetch_features()

        if self.index < len(self.features):
            self.index += 1
            return self.features[self.index - 1]

        raise StopIteration

    def __len__(self):
        """ Length of iterator is number of features which can be obtained from Geopedia with applied filters
        """
        return self.get_size()

    def _fetch_features(self):
        """ Retrieves a new page of features from Geopedia
        """
        if self.next_page_url is None:
            return

        response = get_json(self.next_page_url, post_values=self.query, headers=self.gpd_session.session_headers)

        self.features.extend(response['features'])
        self.next_page_url = response['pagination']['next']
        self.layer_size = response['pagination']['total']

    def get_geometry_iterator(self):
        """ Iterator over Geopedia feature geometries
        """
        for feature in self:
            yield geo_shape(feature['geometry'])

    def get_field_iterator(self, field):
        """ Iterator over the specified field of Geopedia features
        """
        for feature in self:
            yield feature['properties'].get(field, [])

    def get_size(self):
        """ Provides number of features which can be obtained. It has to fetch at least one feature from
        Geopedia services to get this information.

        :return: Size of Geopedia layer with applied filters
        :rtype: int
        """
        if self.layer_size is None:
            self._fetch_features()
        return self.layer_size



"""
Main module for collecting data
"""

import datetime
import os
import logging
import copy
from abc import ABC, abstractmethod

from .ogc import OgcImageService
from .fis import FisService
from .geopedia import GeopediaWmsService, GeopediaImageService
from .aws import AwsProduct, AwsTile
from .aws_safe import SafeProduct, SafeTile
from .download import DownloadClient, AwsDownloadClient, SentinelHubDownloadClient
from .exceptions import DownloadFailedException
from .io_utils import read_data
from .os_utils import make_folder
from .constants import DataSource, MimeType, CustomUrlParam, ServiceType, CRS, HistogramType

LOGGER = logging.getLogger(__name__)


class DataRequest(ABC):
    """ Abstract class for all Sentinel Hub data requests.

    Every data request type can write the fetched data to disk and then read it again (and hence avoid the need to
    download the same data again).

    :param data_folder: location of the directory where the fetched data will be saved.
    :type data_folder: str
    """
    def __init__(self, download_client_class, *, data_folder=None):
        self.download_client_class = download_client_class
        self.data_folder = data_folder.rstrip('/') if data_folder else None

        self.download_list = []
        self.folder_list = []
        self.create_request()

    @abstractmethod
    def create_request(self):
        """ An abstract method for logic of creating download requests
        """
        raise NotImplementedError

    def get_download_list(self):
        """
        Returns a list of download requests for requested data.

        :return: List of data to be downloaded
        :rtype: list(sentinelhub.DownloadRequest)
        """
        return self.download_list

    def get_filename_list(self):
        """
        Returns a list of file names where the requested data will be saved or read from, if it
        was already downloaded and saved.

        :return: List of filenames where downloaded data will be saved.
        :rtype: list(str)
        """
        return [request.filename for request in self.download_list]

    def get_url_list(self):
        """
        Returns a list of urls for requested data.

        :return: List of URLs from where data will be downloaded.
        :rtype: list(str)
        """
        return [request.url for request in self.download_list]

    def is_valid_request(self):
        """ Checks if initialized class instance successfully prepared a list of items to download

        :return: `True` if request is valid and `False` otherwise
        :rtype: bool
        """
        return isinstance(self.download_list, list)

    def get_data(self, *, save_data=False, data_filter=None, redownload=False, max_threads=None,
                 raise_download_errors=True):
        """ Get requested data either by downloading it or by reading it from the disk (if it
        was previously downloaded and saved).

        :param save_data: flag to turn on/off saving of data to disk. Default is `False`.
        :type save_data: bool
        :param redownload: if `True`, download again the requested data even though it's already saved to disk.
                            Default is `False`, do not download if data is already available on disk.
        :type redownload: bool
        :param data_filter: Used to specify which items will be returned by the method and in which order. E.g. with
            ``data_filter=[0, 2, -1]`` the method will return only 1st, 3rd and last item. Default filter is `None`.
        :type data_filter: list(int) or None
        :param max_threads: number of threads to use when downloading data; default is ``max_threads=None`` which
            by default uses the number of processors on the system multiplied by 5.
        :type max_threads: int
        :param raise_download_errors: If `True` any error in download process should be raised as
            ``DownloadFailedException``. If `False` failed downloads will only raise warnings and the method will
            return list with `None` values in places where the results of failed download requests should be.
        :type raise_download_errors: bool
        :return: requested images as numpy arrays, where each array corresponds to a single acquisition and has
                    shape ``[height, width, channels]``.
        :rtype: list of numpy arrays
        """
        self._preprocess_request(save_data, True)
        data_list = self._execute_data_download(data_filter, redownload, max_threads, raise_download_errors)
        return self._add_saved_data(data_list, data_filter, raise_download_errors)

    def save_data(self, *, data_filter=None, redownload=False, max_threads=None, raise_download_errors=False):
        """ Saves data to disk. If ``redownload=True`` then the data is redownloaded using ``max_threads`` workers.

        :param data_filter: Used to specify which items will be returned by the method and in which order. E.g. with
            `data_filter=[0, 2, -1]` the method will return only 1st, 3rd and last item. Default filter is `None`.
        :type data_filter: list(int) or None
        :param redownload: data is redownloaded if ``redownload=True``. Default is `False`
        :type redownload: bool
        :param max_threads: number of threads to use when downloading data; default is ``max_threads=None`` which
            by default uses the number of processors on the system multiplied by 5.
        :type max_threads: int
        :param raise_download_errors: If `True` any error in download process should be raised as
            ``DownloadFailedException``. If `False` failed downloads will only raise warnings.
        :type raise_download_errors: bool
        """
        self._preprocess_request(True, False)
        self._execute_data_download(data_filter, redownload, max_threads, raise_download_errors)

    def _execute_data_download(self, data_filter, redownload, max_threads, raise_download_errors):
        """ Calls download module and executes the download process

        :param data_filter: Used to specify which items will be returned by the method and in which order. E.g. with
            `data_filter=[0, 2, -1]` the method will return only 1st, 3rd and last item. Default filter is `None`.
        :type data_filter: list(int) or None
        :param redownload: data is redownloaded if ``redownload=True``. Default is `False`
        :type redownload: bool
        :param max_threads: the number of workers to use when downloading, default ``max_threads=None``
        :type max_threads: int
        :param raise_download_errors: If `True` any error in download process should be raised as
            ``DownloadFailedException``. If `False` failed downloads will only raise warnings.
        :type raise_download_errors: bool
        :return: List of data obtained from download
        :rtype: list
        """
        is_repeating_filter = False
        if data_filter is None:
            filtered_download_list = self.download_list
        elif isinstance(data_filter, (list, tuple)):
            try:
                filtered_download_list = [self.download_list[index] for index in data_filter]
            except IndexError:
                raise IndexError('Indices of data_filter are out of range')

            filtered_download_list, mapping_list = self._filter_repeating_items(filtered_download_list)
            is_repeating_filter = len(filtered_download_list) < len(mapping_list)
        else:
            raise ValueError('data_filter parameter must be a list of indices')

        client = self.download_client_class(redownload=redownload, raise_download_errors=raise_download_errors)
        data_list = client.download(filtered_download_list, max_threads=max_threads)

        if is_repeating_filter:
            data_list = [copy.deepcopy(data_list[index]) for index in mapping_list]

        return data_list

    @staticmethod
    def _filter_repeating_items(download_list):
        """ Because of data_filter some requests in download list might be the same. In order not to download them again
        this method will reduce the list of requests. It will also return a mapping list which can be used to
        reconstruct the previous list of download requests.

        :param download_list: List of download requests
        :type download_list: list(sentinelhub.DownloadRequest)
        :return: reduced download list with unique requests and mapping list
        :rtype: (list(sentinelhub.DownloadRequest), list(int))
        """
        unique_requests_map = {}
        mapping_list = []
        unique_download_list = []
        for download_request in download_list:
            if download_request not in unique_requests_map:
                unique_requests_map[download_request] = len(unique_download_list)
                unique_download_list.append(download_request)
            mapping_list.append(unique_requests_map[download_request])
        return unique_download_list, mapping_list

    def _preprocess_request(self, save_data, return_data):
        """ Prepares requests for download and creates empty folders

        :param save_data: Tells whether to save data or not
        :type save_data: bool
        :param return_data: Tells whether to return data or not
        :type return_data: bool
        """
        if not self.is_valid_request():
            raise ValueError('Cannot obtain data because request is invalid')

        if save_data and self.data_folder is None:
            raise ValueError('Request parameter `data_folder` is not specified. '
                             'In order to save data please set `data_folder` to location on your disk.')

        for download_request in self.download_list:
            download_request.save_response = save_data
            download_request.return_data = return_data
            download_request.data_folder = self.data_folder

        if save_data:
            for folder in self.folder_list:
                make_folder(os.path.join(self.data_folder, folder))

    def _add_saved_data(self, data_list, data_filter, raise_download_errors):
        """ Adds already saved data that was not redownloaded to the requested data list.
        """
        filtered_download_list = self.download_list if data_filter is None else \
            [self.download_list[index] for index in data_filter]
        for i, request in enumerate(filtered_download_list):
            if request.return_data and data_list[i] is None:
                if os.path.exists(request.get_file_path()):
                    data_list[i] = read_data(request.get_file_path())
                elif raise_download_errors:
                    raise DownloadFailedException('Failed to download data from {}.\n No previously downloaded data '
                                                  'exists in file {}.'.format(request.url, request.get_file_path()))
        return data_list


class OgcRequest(DataRequest):
    """ The base class for OGC-type requests (WMS and WCS) where all common parameters are
    defined.

    :param data_source: Source of requested satellite data. It has to be the same as defined in Sentinel Hub
        Configurator for the given layer. Default is Sentinel-2 L1C.
    :type data_source: constants.DataSource
    :param service_type: type of OGC service (WMS or WCS)
    :type service_type: constants.ServiceType
    :param size_x: number of pixels in x or resolution in x (i.e. ``512`` or ``10m``)
    :type size_x: int or str
    :param size_y: number of pixels in x or resolution in y (i.e. ``512`` or ``10m``)
    :type size_y: int or str
    :param bbox: Bounding box of the requested image. Coordinates must be in the specified coordinate reference system.
    :type bbox: geometry.BBox
    :param time: time or time range for which to return the results, in ISO8601 format
                (year-month-date, for example: ``2016-01-01``, or year-month-dateThours:minuts:seconds format,
                i.e. ``2016-01-01T16:31:21``). When a single time is specified the request will return
                data for that specific date, if it exists. If a time range is specified the result is a list of all
                scenes between the specified dates conforming to the cloud coverage criteria. Most recent acquisition
                being first in the list. For the latest acquisition use ``latest``.
                Examples: ``latest``, ``'2016-01-01'``, or ``('2016-01-01', ' 2016-01-31')``
    :type time: str or (str, str) or datetime.date or (datetime.date, datetime.date) or datetime.datetime or
        (datetime.datetime, datetime.datetime)
    :param layer: An ID of a layer configured in Sentinel Hub Configurator. It has to be configured for the same
        instance ID which will be used for this request. Also the satellite source of the layer in Configurator must
        match the one given by `data_source` parameter
    :type layer: str
    :param maxcc: maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is ``1.0``.
    :type maxcc: float
    :param image_format: format of the returned image by the Sentinel Hub's WMS getMap service. Default is PNG, but
                        in some cases 32-bit TIFF is required, i.e. if requesting unprocessed raw bands.
                        Default is ``constants.MimeType.PNG``.
    :type image_format: constants.MimeType
    :param instance_id: user's instance id. If `None` the instance id is taken from the ``config.json``
                        configuration file.
    :type instance_id: str
    :param custom_url_params: dictionary of CustomUrlParameters and their values supported by Sentinel Hub's WMS and WCS
                              services. All available parameters are described at
                              http://www.sentinel-hub.com/develop/documentation/api/custom-url-parameters. Note: in
                              case of constants.CustomUrlParam.EVALSCRIPT the dictionary value must be a string
                              of Javascript code that is not encoded into base64.
    :type custom_url_params: dictionary of CustomUrlParameter enum and its value, i.e.
                              ``{constants.CustomUrlParam.ATMFILTER:'ATMCOR'}``
    :param time_difference: The time difference below which dates are deemed equal. That is, if for the given set of OGC
                            parameters the images are available at datestimes `d1<=d2<=...<=dn` then only those with
                            `dk-dj>time_difference` will be considered. The default time difference is negative (`-1s`),
                            meaning that all dates are considered by default.
    :type time_difference: datetime.timedelta
    :param data_folder: location of the directory where the fetched data will be saved.
    :type data_folder: str
    """
    def __init__(self, layer, bbox, *, time='latest', service_type=None, data_source=DataSource.SENTINEL2_L1C,
                 size_x=None, size_y=None, maxcc=1.0, image_format=MimeType.PNG, instance_id=None,
                 custom_url_params=None, time_difference=datetime.timedelta(seconds=-1), **kwargs):
        self.layer = layer
        self.bbox = bbox
        self.time = time
        self.data_source = data_source
        self.maxcc = maxcc
        self.image_format = MimeType(image_format)
        self.instance_id = instance_id
        self.service_type = service_type
        self.size_x = size_x
        self.size_y = size_y
        self.custom_url_params = custom_url_params
        self.time_difference = time_difference

        if self.custom_url_params is not None:
            self._check_custom_url_parameters()

        self.wfs_iterator = None

        super().__init__(SentinelHubDownloadClient, **kwargs)

    def _check_custom_url_parameters(self):
        """ Checks if custom url parameters are valid parameters.

        Throws ValueError if the provided parameter is not a valid parameter.
        """
        for param in self.custom_url_params:
            if param not in CustomUrlParam:
                raise ValueError('Parameter %s is not a valid custom url parameter. Please check and fix.' % param)

        if self.service_type is ServiceType.FIS and CustomUrlParam.GEOMETRY in self.custom_url_params:
            raise ValueError('{} should not be a custom url parameter of a FIS request'.format(CustomUrlParam.GEOMETRY))

    def create_request(self, reset_wfs_iterator=False):
        """ Set download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.

        :param reset_wfs_iterator: When re-running the method this flag is used to reset/keep existing ``wfs_iterator``
            (i.e. instance of ``WebFeatureService`` class). If the iterator is not reset you don't have to repeat a
            service call but tiles and dates will stay the same.
        :type reset_wfs_iterator: bool
        """
        if reset_wfs_iterator:
            self.wfs_iterator = None

        ogc_service = OgcImageService(instance_id=self.instance_id)
        self.download_list = ogc_service.get_request(self)
        self.wfs_iterator = ogc_service.get_wfs_iterator()

    def get_dates(self):
        """ Get list of dates

        List of all available Sentinel-2 acquisitions for given bbox with max cloud coverage and the specified
        time interval. When a single time is specified the request will return that specific date, if it exists.
        If a time range is specified the result is a list of all scenes between the specified dates conforming to
        the cloud coverage criteria. Most recent acquisition being first in the list.

        :return: list of all available Sentinel-2 acquisition times within request's time interval and
                acceptable cloud coverage.
        :rtype: list(datetime.datetime) or [None]
        """
        return OgcImageService(instance_id=self.instance_id).get_dates(self)

    def get_tiles(self):
        """ Returns iterator over info about all satellite tiles used for the OgcRequest

        :return: Iterator of dictionaries containing info about all satellite tiles used in the request. In case of
                 DataSource.DEM it returns None.
        :rtype: Iterator[dict] or None
        """
        return self.wfs_iterator


class WmsRequest(OgcRequest):
    """ Web Map Service request class

    Creates an instance of Sentinel Hub WMS (Web Map Service) GetMap request,
    which provides access to Sentinel-2's unprocessed bands (B01, B02, ..., B08, B8A, ..., B12)
    or processed products such as true color imagery, NDVI, etc. The only difference is that in
    the case od WMS request the user specifies the desired image size instead of its resolution.

    It is required to specify at least one of `width` and `height` parameters. If only one of them is specified the
    the other one will be calculated to best fit the bounding box ratio. If both of them are specified they will be used
    no matter the bounding box ratio.

    :param width: width (number of columns) of the returned image (array)
    :type width: int or None
    :param height: height (number of rows) of the returned image (array)
    :type height: int or None
    :param data_source: Source of requested satellite data. It has to be the same as defined in Sentinel Hub
        Configurator for the given layer. Default is Sentinel-2 L1C.
    :type data_source: constants.DataSource
    :param bbox: Bounding box of the requested image. Coordinates must be in the specified coordinate reference system.
    :type bbox: geometry.BBox
    :param time: time or time range for which to return the results, in ISO8601 format
                (year-month-date, for example: ``2016-01-01``, or year-month-dateThours:minuts:seconds format,
                i.e. ``2016-01-01T16:31:21``). When a single time is specified the request will return
                data for that specific date, if it exists. If a time range is specified the result is a list of all
                scenes between the specified dates conforming to the cloud coverage criteria. Most recent acquisition
                being first in the list. For the latest acquisition use ``latest``.
                Examples: ``latest``, ``'2016-01-01'``, or ``('2016-01-01', ' 2016-01-31')``
    :type time: str or (str, str) or datetime.date or (datetime.date, datetime.date) or datetime.datetime or
        (datetime.datetime, datetime.datetime)
    :param layer: An ID of a layer configured in Sentinel Hub Configurator. It has to be configured for the same
        instance ID which will be used for this request. Also the satellite source of the layer in Configurator must
        match the one given by `data_source` parameter
    :type layer: str
    :param maxcc: maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is ``1.0``.
    :type maxcc: float
    :param image_format: format of the returned image by the Sentinel Hub's WMS getMap service. Default is PNG, but
                        in some cases 32-bit TIFF is required, i.e. if requesting unprocessed raw bands.
                        Default is ``constants.MimeType.PNG``.
    :type image_format: constants.MimeType
    :param instance_id: User's Sentinel Hub instance id. If `None` the instance id is taken from the ``config.json``
                        configuration file.
    :type instance_id: str
    :param custom_url_params: dictionary of CustomUrlParameters and their values supported by Sentinel Hub's WMS and WCS
                              services. All available parameters are described at
                              http://www.sentinel-hub.com/develop/documentation/api/custom-url-parameters. Note: in
                              case of constants.CustomUrlParam.EVALSCRIPT the dictionary value must be a string
                              of Javascript code that is not encoded into base64.
    :type custom_url_params: dictionary of CustomUrlParameter enum and its value, i.e.
                              ``{constants.CustomUrlParam.ATMFILTER:'ATMCOR'}``
    :param time_difference: The time difference below which dates are deemed equal. That is, if for the given set of OGC
                            parameters the images are available at datestimes `d1<=d2<=...<=dn` then only those with
                            `dk-dj>time_difference` will be considered. The default time difference is negative (`-1s`),
                            meaning that all dates are considered by default.
    :type time_difference: datetime.timedelta
    :param data_folder: location of the directory where the fetched data will be saved.
    :type data_folder: str

    More info available at:
    https://www.sentinel-hub.com/develop/documentation/api/ogc_api/wms-parameters
    """
    def __init__(self, *, width=None, height=None, **kwargs):
        super().__init__(service_type=ServiceType.WMS, size_x=width, size_y=height, **kwargs)


class WcsRequest(OgcRequest):
    """ Web Coverage Service request class

    Creates an instance of Sentinel Hub WCS (Web Coverage Service) GetCoverage request,
    which provides access to Sentinel-2's unprocessed bands (B01, B02, ..., B08, B8A, ..., B12)
    or processed products such as true color imagery, NDVI, etc., as the WMS service. The
    only difference is that in the case od WCS request the user specifies the desired
    resolution of the image instead of its size.

    More info available at:
    https://www.sentinel-hub.com/develop/documentation/api/ogc_api/wcs-request

    :param resx: resolution in x (resolution of a column) given in meters in the format (examples ``10m``,
                 ``20m``, ...). Default is ``10m``, which is the best native resolution of some Sentinel-2 bands.
    :type resx: str
    :param resy: resolution in y (resolution of a row) given in meters in the format (examples ``10m``, ``20m``, ...).
                Default is ``10m``, which is the best native resolution of some Sentinel-2 bands.
    :type resy: str
    :param data_source: Source of requested satellite data. It has to be the same as defined in Sentinel Hub
        Configurator for the given layer. Default is Sentinel-2 L1C.
    :type data_source: constants.DataSource
    :param bbox: Bounding box of the requested image. Coordinates must be in the specified coordinate reference system.
    :type bbox: geometry.BBox
    :param time: time or time range for which to return the results, in ISO8601 format
                (year-month-date, for example: ``2016-01-01``, or year-month-dateThours:minuts:seconds format,
                i.e. ``2016-01-01T16:31:21``). When a single time is specified the request will return
                data for that specific date, if it exists. If a time range is specified the result is a list of all
                scenes between the specified dates conforming to the cloud coverage criteria. Most recent acquisition
                being first in the list. For the latest acquisition use ``latest``.
                Examples: ``latest``, ``'2016-01-01'``, or ``('2016-01-01', ' 2016-01-31')``
    :type time: str or (str, str) or datetime.date or (datetime.date, datetime.date) or datetime.datetime or
        (datetime.datetime, datetime.datetime)
    :param layer: An ID of a layer configured in Sentinel Hub Configurator. It has to be configured for the same
        instance ID which will be used for this request. Also the satellite source of the layer in Configurator must
        match the one given by `data_source` parameter
    :type layer: str
    :param maxcc: maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is ``1.0``.
    :type maxcc: float
    :param image_format: format of the returned image by the Sentinel Hub's WMS getMap service. Default is PNG, but
                        in some cases 32-bit TIFF is required, i.e. if requesting unprocessed raw bands.
                        Default is ``constants.MimeType.PNG``.
    :type image_format: constants.MimeType
    :param instance_id: user's instance id. If `None` the instance id is taken from the ``config.json``
                        configuration file.
    :type instance_id: str
    :param custom_url_params: dictionary of CustomUrlParameters and their values supported by Sentinel Hub's WMS and WCS
                              services. All available parameters are described at
                              http://www.sentinel-hub.com/develop/documentation/api/custom-url-parameters. Note: in
                              case of constants.CustomUrlParam.EVALSCRIPT the dictionary value must be a string
                              of Javascript code that is not encoded into base64.
    :type custom_url_params: Dictionary of CustomUrlParameter enum and its value, i.e.
                              ``{constants.CustomUrlParam.ATMFILTER:'ATMCOR'}``
    :param time_difference: The time difference below which dates are deemed equal. That is, if for the given set of OGC
                            parameters the images are available at datestimes `d1<=d2<=...<=dn` then only those with
                            `dk-dj>time_difference` will be considered. The default time difference is negative (`-1s`),
                            meaning that all dates are considered by default.
    :type time_difference: datetime.timedelta
    :param data_folder: location of the directory where the fetched data will be saved.
    :type data_folder: str
    """
    def __init__(self, *, resx='10m', resy='10m', **kwargs):
        super().__init__(service_type=ServiceType.WCS, size_x=resx, size_y=resy, **kwargs)


class FisRequest(OgcRequest):
    """ The Statistical info (or feature info service, abbreviated FIS) request class

    The Statistical info (or feature info service, abbreviated FIS), performs elementary statistical
    computations---such as mean, standard deviation, and histogram approximating the distribution of reflectance
    values---on remotely sensed data for a region specified in a given spatial reference system across different
    bands and time ranges.

    A quintessential usage example would be querying the service for basic statistics and the distribution of NDVI
    values for a polygon representing an agricultural unit over a time range.

    More info available at:
    https://www.sentinel-hub.com/develop/documentation/api/ogc_api/wcs-request

    :param layer: An ID of a layer configured in Sentinel Hub Configurator. It has to be configured for the same
        instance ID which will be used for this request. Also the satellite source of the layer in Configurator must
        match the one given by `data_source` parameter
    :type layer: str
    :param time: time or time range for which to return the results, in ISO8601 format
        (year-month-date, for example: ``2016-01-01``, or year-month-dateThours:minuts:seconds format,
        i.e. ``2016-01-01T16:31:21``). Examples: ``'2016-01-01'``, or ``('2016-01-01', ' 2016-01-31')``
    :type time: str or (str, str) or datetime.date or (datetime.date, datetime.date) or datetime.datetime or
        (datetime.datetime, datetime.datetime)
    :param geometry_list: A WKT representation of a geometry describing the region of interest.
        Note that WCS 1.1.1 standard is used here, so for EPSG:4326 coordinates should be in latitude/longitude order.
    :type geometry_list: list, [geometry.Geometry or geometry.Bbox]
    :param resolution: Specifies the spatial resolution, in meters per pixel, of the image from which the statistics
        are to be estimated. When using CRS=EPSG:4326 one has to add the "m" suffix to
        enforce resolution in meters per pixel (e.g. RESOLUTION=10m).
    :type resolution: str
    :param bins: The number of bins (a positive integer) in the histogram. If this parameter is absent no histogram
        is computed.
    :type bins: str
    :param histogram_type: type of histogram
    :type histogram_type: HistogramType
    :param data_source: Source of requested satellite data. It has to be the same as defined in Sentinel Hub
        Configurator for the given layer. Default is Sentinel-2 L1C.
    :type data_source: constants.DataSource
    :param maxcc: maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is ``1.0``.
    :type maxcc: float
    :param instance_id: user's instance id. If `None` the instance id is taken from the ``config.json``
        configuration file.
    :type instance_id: str
    :param custom_url_params: Dictionary of CustomUrlParameters and their values supported by Sentinel Hub's WMS and WCS
        services. All available parameters are described at
        http://www.sentinel-hub.com/develop/documentation/api/custom-url-parameters. Note: in
        case of constants.CustomUrlParam.EVALSCRIPT the dictionary value must be a string
        of Javascript code that is not encoded into base64.
    :type custom_url_params: dictionary of CustomUrlParameter enum and its value, i.e.
        ``{constants.CustomUrlParam.ATMFILTER:'ATMCOR'}``
    :param data_folder: location of the directory where the fetched data will be saved.
    :type data_folder: str
    """
    def __init__(self, layer, time, geometry_list, *, resolution='10m', bins=None, histogram_type=None, **kwargs):
        self.geometry_list = geometry_list
        self.resolution = resolution
        self.bins = bins
        self.histogram_type = HistogramType(histogram_type) if histogram_type else None

        super().__init__(bbox=None, layer=layer, time=time, service_type=ServiceType.FIS, **kwargs)

    def create_request(self):
        """ Set download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.
        """
        fis_service = FisService(instance_id=self.instance_id)
        self.download_list = fis_service.get_request(self)

    def get_dates(self):
        """ This method is not supported for FIS request
        """
        raise NotImplementedError

    def get_tiles(self):
        """ This method is not supported for FIS request
        """
        raise NotImplementedError


class GeopediaRequest(DataRequest):
    """ The base class for Geopedia requests where all common parameters are defined.

    :param layer: Geopedia layer which contains requested data
    :type layer: str
    :param service_type: Type of the service, supported are ``ServiceType.WMS`` and ``ServiceType.IMAGE``
    :type service_type: constants.ServiceType
    :param bbox: Bounding box of the requested data
    :type bbox: geometry.BBox
    :param theme: Geopedia's theme endpoint string for which the layer is defined. Only required by WMS service.
    :type theme: str
    :param image_format: Format of the returned image by the Sentinel Hub's WMS getMap service. Default is
        ``constants.MimeType.PNG``.
    :type image_format: constants.MimeType
    :param data_folder: Location of the directory where the fetched data will be saved.
    :type data_folder: str
    """
    def __init__(self, layer, service_type, *, bbox=None, theme=None, image_format=MimeType.PNG, **kwargs):
        self.layer = layer
        self.service_type = service_type

        self.bbox = bbox
        if bbox.crs is not CRS.POP_WEB:
            raise ValueError('Geopedia Request at the moment supports only bounding boxes with coordinates in '
                             '{}'.format(CRS.POP_WEB))

        self.theme = theme
        self.image_format = MimeType(image_format)

        super().__init__(DownloadClient, **kwargs)

    @abstractmethod
    def create_request(self):
        raise NotImplementedError


class GeopediaWmsRequest(GeopediaRequest):
    """ Web Map Service request class for Geopedia

    Creates an instance of Geopedia's WMS (Web Map Service) GetMap request, which provides access to WMS layers in
    Geopedia.

    :param layer: Geopedia layer which contains requested data
    :type layer: str
    :param theme: Geopedia's theme endpoint string for which the layer is defined.
    :type theme: str
    :param bbox: Bounding box of the requested data
    :type bbox: geometry.BBox
    :param width: width (number of columns) of the returned image (array)
    :type width: int or None
    :param height: height (number of rows) of the returned image (array)
    :type height: int or None
    :param custom_url_params: dictionary of CustomUrlParameters and their values supported by Geopedia's WMS services.
                              At the moment only the transparency is supported (CustomUrlParam.TRANSPARENT).
    :type custom_url_params: dictionary of CustomUrlParameter enum and its value, i.e.
                              ``{constants.CustomUrlParam.TRANSPARENT:True}``
    :param image_format: Format of the returned image by the Sentinel Hub's WMS getMap service. Default is
        ``constants.MimeType.PNG``.
    :type image_format: constants.MimeType
    :param data_folder: Location of the directory where the fetched data will be saved.
    :type data_folder: str
    """
    def __init__(self, layer, theme, bbox, *, width=None, height=None, custom_url_params=None, **kwargs):
        self.size_x = width
        self.size_y = height

        self.custom_url_params = custom_url_params

        if self.custom_url_params is not None:
            self._check_custom_url_parameters()

        super().__init__(layer=layer, theme=theme, bbox=bbox, service_type=ServiceType.WMS, **kwargs)

    def _check_custom_url_parameters(self):
        """ Checks if custom url parameters are valid parameters.

        Throws ValueError if the provided parameter is not a valid parameter.
        """
        for param in self.custom_url_params.keys():
            if param is not CustomUrlParam.TRANSPARENT:
                raise ValueError('Parameter {} is currently not supported.'.format(param))

    def create_request(self):
        """ Set download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.
        """
        gpd_service = GeopediaWmsService()
        self.download_list = gpd_service.get_request(self)


class GeopediaImageRequest(GeopediaRequest):
    """ Request to access data in a Geopedia vector / raster layer.

    :param image_field_name: Name of the field in the data table which holds images
    :type image_field_name: str
    :param keep_image_names: If `True` images will be saved with the same names as in Geopedia otherwise Geopedia
        hashes will be used as names. If there are multiple images with the same names in the Geopedia layer this
        parameter should be set to `False` to prevent images being overwritten.
    :type keep_image_names: bool
    :param layer: Geopedia layer which contains requested data
    :type layer: str
    :param bbox: Bounding box of the requested data
    :type bbox: geometry.BBox
    :param image_format: Format of the returned image by the Sentinel Hub's WMS getMap service. Default is
        ``constants.MimeType.PNG``.
    :type image_format: constants.MimeType
    :param gpd_session: Optional parameter for specifying a custom Geopedia session, which can also contain login
        credentials. This can be used for accessing private Geopedia layers. By default it is set to `None` and a basic
        Geopedia session without credentials will be created.
    :type gpd_session: GeopediaSession or None
    :param data_folder: Location of the directory where the fetched data will be saved.
    :type data_folder: str
    """
    def __init__(self, *, image_field_name, keep_image_names=True, gpd_session=None, **kwargs):
        self.image_field_name = image_field_name
        self.keep_image_names = keep_image_names
        self.gpd_session = gpd_session

        self.gpd_iterator = None

        super().__init__(service_type=ServiceType.IMAGE, **kwargs)

    def create_request(self, reset_gpd_iterator=False):
        """ Set a list of download requests

        Set a list of DownloadRequests for all images that are under the
        given property of the Geopedia's Vector layer.

        :param reset_gpd_iterator: When re-running the method this flag is used to reset/keep existing ``gpd_iterator``
            (i.e. instance of ``GeopediaFeatureIterator`` class). If the iterator is not reset you don't have to
            repeat a service call but tiles and dates will stay the same.
        :type reset_gpd_iterator: bool
        """
        if reset_gpd_iterator:
            self.gpd_iterator = None

        gpd_service = GeopediaImageService()
        self.download_list = gpd_service.get_request(self)
        self.gpd_iterator = gpd_service.get_gpd_iterator()

    def get_items(self):
        """ Returns iterator over info about data used for this request

        :return: Iterator of dictionaries containing info about data used in
                 this request.
        :rtype: Iterator[dict] or None
        """
        return self.gpd_iterator


class AwsRequest(DataRequest):
    """ The base class for Amazon Web Service request classes. Common parameters are defined here.

    Collects and provides data from AWS.

    AWS database is available at:
    http://sentinel-s2-l1c.s3-website.eu-central-1.amazonaws.com/

    :param bands: List of Sentinel-2 bands for request. If `None` all bands will be obtained
    :type bands: list(str) or None
    :param metafiles: list of additional metafiles available on AWS
                      (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``)
    :type metafiles: list(str)
    :param safe_format: flag that determines the structure of saved data. If `True` it will be saved in .SAFE format
                        defined by ESA. If `False` it will be saved in the same structure as the stucture at AWS.
    :type safe_format: bool
    :param data_folder: location of the directory where the fetched data will be saved.
    :type data_folder: str
    """
    def __init__(self, *, bands=None, metafiles=None, safe_format=False, **kwargs):
        self.bands = bands
        self.metafiles = metafiles
        self.safe_format = safe_format

        self.aws_service = None
        super().__init__(AwsDownloadClient, **kwargs)

    @abstractmethod
    def create_request(self):
        raise NotImplementedError

    def get_aws_service(self):
        """
        :return: initialized AWS service class
        :rtype: aws.AwsProduct or aws.AwsTile or aws_safe.SafeProduct or aws_safe.SafeTile
        """
        return self.aws_service


class AwsProductRequest(AwsRequest):
    """ AWS Service request class for an ESA product

    List of available products:
    http://sentinel-s2-l1c.s3-website.eu-central-1.amazonaws.com/#products/

    :param product_id: original ESA product identification string
                       (e.g. ``'S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551'``)
    :type product_id: str
    :param tile_list: list of tiles inside the product to be downloaded. If parameter is set to `None` all
                      tiles inside the product will be downloaded.
    :type tile_list: list(str) or None
    :param bands: List of Sentinel-2 bands for request. If `None` all bands will be obtained
    :type bands: list(str) or None
    :param metafiles: list of additional metafiles available on AWS
                      (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``)
    :type metafiles: list(str)
    :param safe_format: flag that determines the structure of saved data. If `True` it will be saved in .SAFE format
                        defined by ESA. If `False` it will be saved in the same structure as the stucture at AWS.
    :type safe_format: bool
    :param data_folder: location of the directory where the fetched data will be saved.
    :type data_folder: str
    """
    def __init__(self, product_id, *, tile_list=None, **kwargs):
        self.product_id = product_id
        self.tile_list = tile_list

        super().__init__(**kwargs)

    def create_request(self):
        if self.safe_format:
            self.aws_service = SafeProduct(self.product_id, tile_list=self.tile_list, bands=self.bands,
                                           metafiles=self.metafiles)
        else:
            self.aws_service = AwsProduct(self.product_id, tile_list=self.tile_list, bands=self.bands,
                                          metafiles=self.metafiles)

        self.download_list, self.folder_list = self.aws_service.get_requests()


class AwsTileRequest(AwsRequest):
    """ AWS Service request class for an ESA tile

    List of available products:
    http://sentinel-s2-l1c.s3-website.eu-central-1.amazonaws.com/#tiles/

    :param tile: tile name (e.g. ``'T10UEV'``)
    :type tile: str
    :param time: tile sensing time in ISO8601 format
    :type time: str
    :param aws_index: there exist Sentinel-2 tiles with the same tile and time parameter. Therefore each tile on AWS
                      also has an index which is visible in their url path. If aws_index is set to `None` the class
                      will try to find the index automatically. If there will be multiple choices it will choose the
                      lowest index and inform the user.
    :type aws_index: int or None
    :param data_source: Source of requested AWS data. Supported sources are Sentinel-2 L1C and Sentinel-2 L2A, default
                        is Sentinel-2 L1C data.
    :type data_source: constants.DataSource
    :param bands: List of Sentinel-2 bands for request. If `None` all bands will be obtained
    :type bands: list(str) or None
    :param metafiles: list of additional metafiles available on AWS
                      (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``)
    :type metafiles: list(str)
    :param safe_format: flag that determines the structure of saved data. If `True` it will be saved in .SAFE format
                        defined by ESA. If `False` it will be saved in the same structure as the stucture at AWS.
    :type safe_format: bool
    :param data_folder: location of the directory where the fetched data will be saved.
    :type data_folder: str
    """
    def __init__(self, *, tile=None, time=None, aws_index=None, data_source=DataSource.SENTINEL2_L1C, **kwargs):
        self.tile = tile
        self.time = time
        self.aws_index = aws_index
        self.data_source = data_source

        super().__init__(**kwargs)

    def create_request(self):
        if self.safe_format:
            self.aws_service = SafeTile(self.tile, self.time, self.aws_index, bands=self.bands,
                                        metafiles=self.metafiles, data_source=self.data_source)
        else:
            self.aws_service = AwsTile(self.tile, self.time, self.aws_index, bands=self.bands,
                                       metafiles=self.metafiles, data_source=self.data_source)

        self.download_list, self.folder_list = self.aws_service.get_requests()


def get_safe_format(product_id=None, tile=None, entire_product=False, bands=None, data_source=DataSource.SENTINEL2_L1C):
    """ Returns .SAFE format structure in form of nested dictionaries. Either ``product_id`` or ``tile`` must be
    specified.

    :param product_id: original ESA product identification string. Default is `None`
    :type product_id: str
    :param tile: tuple containing tile name and sensing time/date. Default is `None`
    :type tile: (str, str)
    :param entire_product: in case tile is specified this flag determines if it will be place inside a .SAFE structure
                           of the product. Default is `False`
    :type entire_product: bool
    :param bands: list of bands to download. If `None` all bands will be downloaded. Default is `None`
    :type bands: list(str) or None
    :param data_source: In case of tile request the source of satellite data has to be specified. Default is Sentinel-2
                        L1C data.
    :type data_source: constants.DataSource
    :return: Nested dictionaries representing .SAFE structure.
    :rtype: dict
    """
    entire_product = entire_product and product_id is None
    if tile is not None:
        safe_tile = SafeTile(tile_name=tile[0], time=tile[1], bands=bands, data_source=data_source)
        if not entire_product:
            return safe_tile.get_safe_struct()
        product_id = safe_tile.get_product_id()
    if product_id is None:
        raise ValueError('Either product_id or tile must be specified')
    safe_product = SafeProduct(product_id, tile_list=[tile[0]], bands=bands) if entire_product else \
        SafeProduct(product_id, bands=bands)
    return safe_product.get_safe_struct()


def download_safe_format(product_id=None, tile=None, folder='.', redownload=False, entire_product=False, bands=None,
                         data_source=DataSource.SENTINEL2_L1C):
    """ Downloads .SAFE format structure in form of nested dictionaries. Either ``product_id`` or ``tile`` must
    be specified.

    :param product_id: original ESA product identification string. Default is `None`
    :type product_id: str
    :param tile: tuple containing tile name and sensing time/date. Default is `None`
    :type tile: (str, str)
    :param folder: location of the directory where the fetched data will be saved. Default is ``'.'``
    :type folder: str
    :param redownload: if `True`, download again the requested data even though it's already saved to disk. If
                       `False`, do not download if data is already available on disk. Default is `False`
    :type redownload: bool
    :param entire_product: in case tile is specified this flag determines if it will be place inside a .SAFE structure
                           of the product. Default is `False`
    :type entire_product: bool
    :param bands: list of bands to download. If `None` all bands will be downloaded. Default is `None`
    :type bands: list(str) or None
    :param data_source: In case of tile request the source of satellite data has to be specified. Default is Sentinel-2
                        L1C data.
    :type data_source: constants.DataSource
    :return: Nested dictionaries representing .SAFE structure.
    :rtype: dict
    """
    entire_product = entire_product and product_id is None
    if tile is not None:
        safe_request = AwsTileRequest(tile=tile[0], time=tile[1], data_folder=folder, bands=bands,
                                      safe_format=True, data_source=data_source)
        if entire_product:
            safe_tile = safe_request.get_aws_service()
            product_id = safe_tile.get_product_id()
    if product_id is not None:
        safe_request = AwsProductRequest(product_id, tile_list=[tile[0]], data_folder=folder, bands=bands,
                                         safe_format=True) if entire_product else \
            AwsProductRequest(product_id, data_folder=folder, bands=bands, safe_format=True)

    safe_request.save_data(redownload=redownload)



"""
Module implementing DownloadRequest class
"""

import os
import warnings

from ..constants import MimeType, RequestType
from ..exceptions import SHRuntimeWarning
from ..os_utils import sys_is_windows


class DownloadRequest:
    """ Class to manage HTTP requests

    Container for all download requests issued by the DataRequests containing
    url to Sentinel Hub's services or other sources to data, file names to
    saved data and other necessary flags needed when the data is downloaded and
    interpreted.

    :param url: url to Sentinel Hub's services or other sources from where the data is downloaded. Default is `None`
    :type url: str
    :param data_folder: folder name where the fetched data will be (or already is) saved. Default is `None`
    :type data_folder: str
    :param filename: filename of the file where the fetched data will be (or already is) saved. Default is `None`
    :type filename: str
    :param headers: add HTTP headers to request. Default is `None`
    :type headers: dict
    :param request_type: type of request, either GET or POST. Default is ``constants.RequestType.GET``
    :type request_type: constants.RequestType
    :param post_values: form encoded data to send in POST request. Default is `None`
    :type post_values: dict
    :param save_response: flag to turn on/off saving data downloaded by this request to disk. Default is `True`.
    :type save_response: bool
    :param return_data: flag to return or not data downloaded by this request to disk. Default is `True`.
    :type return_data: bool
    :param data_type: expected file format of downloaded data. Default is ``constants.MimeType.RAW``
    :type data_type: constants.MimeType
    """

    def __init__(self, *, url=None, data_folder=None, filename=None, headers=None, request_type=RequestType.GET,
                 post_values=None, save_response=False, return_data=True, data_type=MimeType.RAW,
                 hash_save=False, **properties):

        if hash_save and data_folder is None:
            raise ValueError("When using hash_save, data_folder must also be set.")
        if hash_save and save_response:
            raise ValueError("Either hash_save or save_response should be set, not both.")

        self.url = url
        self.data_folder = data_folder
        self.filename = filename
        self.headers = {} if headers is None else headers
        self.post_values = post_values
        self.save_response = save_response
        self.return_data = return_data

        self.properties = properties

        self.request_type = RequestType(request_type)
        self.data_type = MimeType(data_type)
        self.hash_save = hash_save
        self.request_path = None

        self.file_path = self._file_path()

    def _file_path(self):
        if self.data_folder and self.filename:
            file_path = os.path.join(self.data_folder, self.filename.lstrip('/'))
        elif self.filename:
            file_path = self.filename
        else:
            file_path = None

        if file_path and len(file_path) > 255 and sys_is_windows():
            message = 'File path {} is longer than 255 character which might cause an error while saving on ' \
                      'disk'.format(self.file_path)
            warnings.warn(message, category=SHRuntimeWarning)

        elif file_path and len(self.filename) > 255:
            message = 'Filename {} is longer than 255 character which might cause an error while saving on ' \
                      'disk'.format(self.filename)
            warnings.warn(message, category=SHRuntimeWarning)

        return file_path

    def raise_if_invalid(self):
        """ Method that raises an error if something is wrong with request parameters

        :raises: ValueError
        """
        if self.save_response and self.data_folder is None:
            raise ValueError('Data folder is not specified. '
                             'Please give a data folder name in the initialization of your request.')



"""
Module for working with large geographical areas
"""

import os
import itertools
from abc import ABC, abstractmethod
import json
import math

import shapely.ops
import shapely.geometry
from shapely.geometry import Polygon, MultiPolygon, GeometryCollection

from .geometry import BBox, BBoxCollection, BaseGeometry, Geometry
from .constants import CRS, DataSource
from .geo_utils import transform_point
from .ogc import WebFeatureService


class AreaSplitter(ABC):
    """ Abstract class for splitter classes. It implements common methods used for splitting large area into smaller
    parts.

    :param shape_list: A list of geometrical shapes describing the area of interest
    :type shape_list: list(shapely.geometry.multipolygon.MultiPolygon or shapely.geometry.polygon.Polygon)
    :param crs: Coordinate reference system of the shapes in `shape_list`
    :type crs: CRS
    :param reduce_bbox_sizes: If `True` it will reduce the sizes of bounding boxes so that they will tightly fit the
        given geometry in `shape_list`.
    :type reduce_bbox_sizes: bool
    """
    def __init__(self, shape_list, crs, reduce_bbox_sizes=False):
        self.crs = CRS(crs)
        self._parse_shape_list(shape_list, self.crs)
        self.shape_list = shape_list
        self.area_shape = self._join_shape_list(shape_list)
        self.reduce_bbox_sizes = reduce_bbox_sizes

        self.area_bbox = self.get_area_bbox()
        self.bbox_list = None
        self.info_list = None

    @staticmethod
    def _parse_shape_list(shape_list, crs):
        """ Checks if the given list of shapes is in correct format and parses geometry objects

        :param shape_list: The parameter `shape_list` from class initialization
        :type shape_list: list(shapely.geometry.multipolygon.MultiPolygon or shapely.geometry.polygon.Polygon)
        :raises: ValueError
        """
        if not isinstance(shape_list, list):
            raise ValueError('Splitter must be initialized with a list of shapes')

        return [AreaSplitter._parse_shape(shape, crs) for shape in shape_list]

    @staticmethod
    def _parse_shape(shape, crs):
        """ Helper method for parsing input shapes
        """
        if isinstance(shape, (Polygon, MultiPolygon)):
            return shape
        if isinstance(shape, BaseGeometry):
            return shape.transform(crs).geometry
        raise ValueError('The list of shapes must contain shapes of types {}, {} or subtype of '
                         '{}'.format(type(Polygon), type(MultiPolygon), type(BaseGeometry)))

    @staticmethod
    def _parse_split_parameters(split_parameter, allow_float=False):
        """ Parses the parameters defining the splitting of the BBox

        :param split_parameter: The parameters defining the split. A tuple of int for `BBoxSplitter`, a tuple of float
            for `BaseUtmSplitter`
        :type split_parameter: int or (int, int) or float or (float, float)
        :param allow_float: Whether to check for floats or not
        :type allow_float: bool
        :return: A tuple of n
        :rtype: (int, int)
        :raises: ValueError
        """
        parameters_type = (int, float) if allow_float else int
        if isinstance(split_parameter, parameters_type):
            return split_parameter, split_parameter
        if isinstance(split_parameter, (tuple, list)):
            if len(split_parameter) == 2 and all(isinstance(param, parameters_type) for param in split_parameter):
                return split_parameter[0], split_parameter[1]
            raise ValueError("Split parameter {} must be 2 int{}.".format(split_parameter,
                                                                          '/float' if allow_float else ''))
        raise ValueError("Split parameter must be an int{0} or a tuple of 2 int{0}.".format('/float' if allow_float
                                                                                            else ''))

    @staticmethod
    def _join_shape_list(shape_list):
        """ Joins a list of shapes together into one shape

        :param shape_list: A list of geometrical shapes describing the area of interest
        :type shape_list: list(shapely.geometry.multipolygon.MultiPolygon or shapely.geometry.polygon.Polygon)
        :return: A multipolygon which is a union of shapes in given list
        :rtype: shapely.geometry.multipolygon.MultiPolygon
        """
        return shapely.ops.cascaded_union(shape_list)

    @abstractmethod
    def _make_split(self):
        """ The abstract method where the splitting will happen
        """
        raise NotImplementedError

    def get_bbox_list(self, crs=None, buffer=None, reduce_bbox_sizes=None):
        """ Returns a list of bounding boxes that are the result of the split

        :param crs: Coordinate reference system in which the bounding boxes should be returned. If `None` the CRS will
            be the default CRS of the splitter.
        :type crs: CRS or None
        :param buffer: A percentage of each BBox size increase. This will cause neighbouring bounding boxes to overlap.
        :type buffer: float or None
        :param reduce_bbox_sizes: If `True` it will reduce the sizes of bounding boxes so that they will tightly
            fit the given geometry in `shape_list`. This overrides the same parameter from constructor
        :type reduce_bbox_sizes: bool
        :return: List of bounding boxes
        :rtype: list(BBox)
        """
        bbox_list = self.bbox_list
        if buffer:
            bbox_list = [bbox.buffer(buffer) for bbox in bbox_list]

        if reduce_bbox_sizes is None:
            reduce_bbox_sizes = self.reduce_bbox_sizes
        if reduce_bbox_sizes:
            bbox_list = self._reduce_sizes(bbox_list)

        if crs:
            return [bbox.transform(crs) for bbox in bbox_list]
        return bbox_list

    def get_geometry_list(self):
        """ For each bounding box an intersection with the shape of entire given area is calculated. CRS of the returned
        shapes is the same as CRS of the given area.

        :return: List of polygons or multipolygons corresponding to the order of bounding boxes
        :rtype: list(shapely.geometry.multipolygon.MultiPolygon or shapely.geometry.polygon.Polygon)
        """
        return [self._intersection_area(bbox) for bbox in self.bbox_list]

    def get_info_list(self):
        """ Returns a list of dictionaries containing information about bounding boxes obtained in split. The order in
        the list matches the order of the list of bounding boxes.

        :return: List of dictionaries
        :rtype: list(BBox)
        """
        return self.info_list

    def get_area_shape(self):
        """ Returns a single shape of entire area described with `shape_list` parameter

        :return: A multipolygon which is a union of shapes describing the area
        :rtype: shapely.geometry.multipolygon.MultiPolygon
        """
        return self.area_shape

    def get_area_bbox(self, crs=None):
        """ Returns a bounding box of the entire area

        :param crs: Coordinate reference system in which the bounding box should be returned. If `None` the CRS will
            be the default CRS of the splitter.
        :type crs: CRS or None
        :return: A bounding box of the area defined by the `shape_list`
        :rtype: BBox
        """
        bbox_list = [BBox(shape.bounds, crs=self.crs) for shape in self.shape_list]
        area_minx = min([bbox.lower_left[0] for bbox in bbox_list])
        area_miny = min([bbox.lower_left[1] for bbox in bbox_list])
        area_maxx = max([bbox.upper_right[0] for bbox in bbox_list])
        area_maxy = max([bbox.upper_right[1] for bbox in bbox_list])
        bbox = BBox([area_minx, area_miny, area_maxx, area_maxy], crs=self.crs)
        if crs is None:
            return bbox
        return bbox.transform(crs)

    def _intersects_area(self, bbox):
        """ Checks if the bounding box intersects the entire area

        :param bbox: A bounding box
        :type bbox: BBox
        :return: `True` if bbox intersects the entire area else False
        :rtype: bool
        """
        return self._bbox_to_area_polygon(bbox).intersects(self.area_shape)

    def _intersection_area(self, bbox):
        """ Calculates the intersection of a given bounding box and the entire area

        :param bbox: A bounding box
        :type bbox: BBox
        :return: A shape of intersection
        :rtype: shapely.geometry.multipolygon.MultiPolygon or shapely.geometry.polygon.Polygon
        """
        return self._bbox_to_area_polygon(bbox).intersection(self.area_shape)

    def _bbox_to_area_polygon(self, bbox):
        """ Transforms bounding box into a polygon object in the area CRS.

        :param bbox: A bounding box
        :type bbox: BBox
        :return: A polygon
        :rtype: shapely.geometry.polygon.Polygon
        """
        projected_bbox = bbox.transform(self.crs)
        return projected_bbox.geometry

    def _reduce_sizes(self, bbox_list):
        """ Reduces sizes of bounding boxes
        """
        return [BBox(self._intersection_area(bbox).bounds, self.crs).transform(bbox.crs) for bbox in bbox_list]


class BBoxSplitter(AreaSplitter):
    """ A tool that splits the given area into smaller parts. Given the area it calculates its bounding box and splits
    it into smaller bounding boxes of equal size. Then it filters out the bounding boxes that do not intersect the
    area. If specified by user it can also reduce the sizes of the remaining bounding boxes to best fit the area.

    :param shape_list: A list of geometrical shapes describing the area of interest
    :type shape_list: list(shapely.geometry.multipolygon.MultiPolygon or shapely.geometry.polygon.Polygon)
    :param crs: Coordinate reference system of the shapes in `shape_list`
    :type crs: CRS
    :param split_shape: Parameter that describes the shape in which the area bounding box will be split. It can be a
                        tuple of the form `(n, m)` which means the area bounding box will be split into `n` columns and
                        `m` rows. It can also be a single integer `n` which is the same as `(n, n)`.
    :type split_shape: int or (int, int)
    :param reduce_bbox_sizes: If `True` it will reduce the sizes of bounding boxes so that they will tightly fit
        the given area geometry from `shape_list`.
    :type reduce_bbox_sizes: bool
    """
    def __init__(self, shape_list, crs, split_shape, **kwargs):
        super().__init__(shape_list, crs, **kwargs)

        self.split_shape = self._parse_split_parameters(split_shape)

        self._make_split()

    def _make_split(self):
        """ This method makes the split
        """
        columns, rows = self.split_shape
        bbox_partition = self.area_bbox.get_partition(num_x=columns, num_y=rows)

        self.bbox_list = []
        self.info_list = []
        for i, j in itertools.product(range(columns), range(rows)):
            if self._intersects_area(bbox_partition[i][j]):
                self.bbox_list.append(bbox_partition[i][j])

                info = {'parent_bbox': self.area_bbox,
                        'index_x': i,
                        'index_y': j}
                self.info_list.append(info)


class OsmSplitter(AreaSplitter):
    """ A tool that splits the given area into smaller parts. For the splitting it uses Open Street Map (OSM) grid on
    the specified zoom level. It calculates bounding boxes of all OSM tiles that intersect the area. If specified by
    user it can also reduce the sizes of the remaining bounding boxes to best fit the area.

    :param shape_list: A list of geometrical shapes describing the area of interest
    :type shape_list: list(shapely.geometry.multipolygon.MultiPolygon or shapely.geometry.polygon.Polygon)
    :param crs: Coordinate reference system of the shapes in `shape_list`
    :type crs: CRS
    :param zoom_level: A zoom level defined by OSM. Level 0 is entire world, level 1 splits the world into 4 parts, etc.
    :type zoom_level: int
    :param reduce_bbox_sizes: If `True` it will reduce the sizes of bounding boxes so that they will tightly fit
        the given area geometry from `shape_list`.
    :type reduce_bbox_sizes: bool
    """
    POP_WEB_MAX = transform_point((180, 0), CRS.WGS84, CRS.POP_WEB)[0]

    def __init__(self, shape_list, crs, zoom_level, **kwargs):
        super().__init__(shape_list, crs, **kwargs)

        self.zoom_level = zoom_level

        self._make_split()

    def _make_split(self, ):
        """This method makes the split
        """
        self.area_bbox = self.get_area_bbox(CRS.POP_WEB)
        self._check_area_bbox()

        self.bbox_list = []
        self.info_list = []
        self._recursive_split(self.get_world_bbox(), 0, 0, 0)

        for i, bbox in enumerate(self.bbox_list):
            self.bbox_list[i] = bbox.transform(self.crs)

    def _check_area_bbox(self):
        """ The method checks if the area bounding box is completely inside the OSM grid. That means that its latitudes
        must be contained in the interval (-85.0511, 85.0511)

        :raises: ValueError
        """
        for coord in self.area_bbox:
            if abs(coord) > self.POP_WEB_MAX:
                raise ValueError('OsmTileSplitter only works for areas which have latitude in interval '
                                 '(-85.0511, 85.0511)')

    def get_world_bbox(self):
        """ Creates a bounding box of the entire world in EPSG: 3857

        :return: Bounding box of entire world
        :rtype: BBox
        """
        return BBox((-self.POP_WEB_MAX, -self.POP_WEB_MAX, self.POP_WEB_MAX, self.POP_WEB_MAX), crs=CRS.POP_WEB)

    def _recursive_split(self, bbox, zoom_level, column, row):
        """ Method that recursively creates bounding boxes of OSM grid that intersect the area.

        :param bbox: Bounding box
        :type bbox: BBox
        :param zoom_level: OSM zoom level
        :type zoom_level: int
        :param column: Column in the OSM grid
        :type column: int
        :param row: Row in the OSM grid
        :type row: int
        """
        if zoom_level == self.zoom_level:
            self.bbox_list.append(bbox)
            self.info_list.append({'zoom_level': zoom_level,
                                   'index_x': column,
                                   'index_y': row})
            return

        bbox_partition = bbox.get_partition(num_x=2, num_y=2)
        for i, j in itertools.product(range(2), range(2)):
            if self._intersects_area(bbox_partition[i][j]):
                self._recursive_split(bbox_partition[i][j], zoom_level + 1, 2 * column + i, 2 * row + 1 - j)


class TileSplitter(AreaSplitter):
    """ A tool that splits the given area into smaller parts. Given the area, time interval and data source it collects
    info from Sentinel Hub WFS service about all satellite tiles intersecting the area. For each of them it calculates
    bounding box and if specified it splits these bounding boxes into smaller bounding boxes. Then it filters out the
    ones that do not intersect the area. If specified by user it can also reduce the sizes of the remaining bounding
    boxes to best fit the area.

    :param shape_list: A list of geometrical shapes describing the area of interest
    :type shape_list: list(shapely.geometry.multipolygon.MultiPolygon or shapely.geometry.polygon.Polygon)
    :param crs: Coordinate reference system of the shapes in `shape_list`
    :type crs: CRS
    :param time_interval: Interval with start and end date of the form YYYY-MM-DDThh:mm:ss or YYYY-MM-DD
    :type time_interval: (str, str)
    :param tile_split_shape: Parameter that describes the shape in which the satellite tile bounding boxes will be
                             split. It can be a tuple of the form `(n, m)` which means the tile bounding boxes will be
                             split into `n` columns and `m` rows. It can also be a single integer `n` which is the same
                             as `(n, n)`.
    :type split_shape: int or (int, int)
    :param data_source: Source of requested satellite data. Default is Sentinel-2 L1C data.
    :type data_source: sentinelhub.constants.DataSource
    :param instance_id: User's Sentinel Hub instance id. If `None` the instance id is taken from the ``config.json``
                        configuration file.
    :type instance_id: str
    :param reduce_bbox_sizes: If `True` it will reduce the sizes of bounding boxes so that they will tightly fit the
        given area geometry from `shape_list`.
    :type reduce_bbox_sizes: bool
    """
    def __init__(self, shape_list, crs, time_interval, tile_split_shape=1, data_source=DataSource.SENTINEL2_L1C,
                 instance_id=None, **kwargs):
        super().__init__(shape_list, crs, **kwargs)

        if data_source is DataSource.DEM:
            raise ValueError('This splitter does not support splitting area by DEM tiles. Please specify some other '
                             'DataSource')

        self.time_interval = time_interval
        self.tile_split_shape = tile_split_shape
        self.data_source = data_source
        self.instance_id = instance_id

        self.tile_dict = None

        self._make_split()

    def _make_split(self):
        """ This method makes the split
        """
        self.tile_dict = {}

        wfs = WebFeatureService(self.area_bbox, self.time_interval, data_source=self.data_source,
                                instance_id=self.instance_id)
        date_list = wfs.get_dates()
        geometry_list = wfs.get_geometries()
        for tile_info, (date, geometry) in zip(wfs, zip(date_list, geometry_list)):
            tile_name = ''.join(tile_info['properties']['path'].split('/')[4:7])
            if tile_name not in self.tile_dict:
                self.tile_dict[tile_name] = {'bbox': BBox(tile_info['properties']['mbr'],
                                                          crs=tile_info['properties']['crs']),
                                             'times': [],
                                             'geometries': []}
            self.tile_dict[tile_name]['times'].append(date)
            self.tile_dict[tile_name]['geometries'].append(geometry)

        self.tile_dict = {tile_name: tile_props for tile_name, tile_props in self.tile_dict.items() if
                          self._intersects_area(tile_props['bbox'])}

        self.bbox_list = []
        self.info_list = []

        for tile_name, tile_info in self.tile_dict.items():
            tile_bbox = tile_info['bbox']
            bbox_splitter = BBoxSplitter([tile_bbox.geometry], tile_bbox.crs,
                                         split_shape=self.tile_split_shape)

            for bbox, info in zip(bbox_splitter.get_bbox_list(), bbox_splitter.get_info_list()):
                if self._intersects_area(bbox):
                    info['tile'] = tile_name

                    self.bbox_list.append(bbox)
                    self.info_list.append(info)

    def get_tile_dict(self):
        """ Returns the dictionary of satellite tiles intersecting the area geometry. For each tile they contain info
        about their bounding box and lists of acquisitions and geometries

        :return: Dictionary containing info about tiles intersecting the area
        :rtype: dict
        """
        return self.tile_dict


class CustomGridSplitter(AreaSplitter):
    """ Splitting class which can split according to given custom collection of bounding boxes

    :param shape_list: A list of geometrical shapes describing the area of interest
    :type shape_list: list(shapely.geometry.multipolygon.MultiPolygon or shapely.geometry.polygon.Polygon)
    :param crs: Coordinate reference system of the shapes in `shape_list`
    :type crs: CRS
    :param bbox_grid: A collection of bounding boxes defining a grid of splitting. All of them have to be in the same
        CRS.
    :type bbox_grid: list(BBox) or BBoxCollection
    :param bbox_split_shape: Parameter that describes the shape in which each of the bounding boxes in the given grid
        will be split. It can be a tuple of the form `(n, m)` which means the tile bounding boxes will be
        split into `n` columns and `m` rows. It can also be a single integer `n` which is the same as `(n, n)`.
    :type bbox_split_shape: int or (int, int)
    :param reduce_bbox_sizes: If `True` it will reduce the sizes of bounding boxes so that they will tightly fit
        the given geometry in `shape_list`.
    :type reduce_bbox_sizes: bool
    """
    def __init__(self, shape_list, crs, bbox_grid, bbox_split_shape=1, **kwargs):
        super().__init__(shape_list, crs, **kwargs)

        self.bbox_grid = self._parse_bbox_grid(bbox_grid)
        self.bbox_split_shape = bbox_split_shape

        self._make_split()

    @staticmethod
    def _parse_bbox_grid(bbox_grid):
        """ Helper method for parsing bounding box grid. It will try to parse it into `BBoxCollection`
        """
        if isinstance(bbox_grid, BBoxCollection):
            return bbox_grid

        if isinstance(bbox_grid, list):
            return BBoxCollection(bbox_grid)

        raise ValueError("Parameter 'bbox_grid' should be an instance of {}".format(BBoxCollection.__name__))

    def _make_split(self):
        """ This method makes the split
        """
        self.bbox_list = []
        self.info_list = []

        for grid_idx, grid_bbox in enumerate(self.bbox_grid):
            if self._intersects_area(grid_bbox):

                bbox_splitter = BBoxSplitter([grid_bbox.geometry], grid_bbox.crs,
                                             split_shape=self.bbox_split_shape)

                for bbox, info in zip(bbox_splitter.get_bbox_list(), bbox_splitter.get_info_list()):
                    if self._intersects_area(bbox):
                        info['grid_index'] = grid_idx

                        self.bbox_list.append(bbox)
                        self.info_list.append(info)


class BaseUtmSplitter(AreaSplitter):
    """ Base splitter that returns bboxes of fixed size aligned to UTM zones or UTM grid tiles as defined by the MGRS
    """
    def __init__(self, shape_list, crs, bbox_size):
        """
        :param shape_list: A list of geometrical shapes describing the area of interest
        :type shape_list: list(shapely.geometry.multipolygon.MultiPolygon or shapely.geometry.polygon.Polygon)
        :param crs: Coordinate reference system of the shapes in `shape_list`
        :type crs: CRS
        :param bbox_size: Physical size in metres of generated bounding boxes. Could be a float or tuple of floats
        :type bbox_size: int or (int, int) or float or (float, float)
        """
        super().__init__(shape_list, crs)

        self.bbox_size = self._parse_split_parameters(bbox_size, allow_float=True)

        self.shape_geometry = Geometry(self.area_shape, self.crs).transform(CRS.WGS84)

        self.utm_grid = self._get_utm_polygons()

        self._make_split()

    @abstractmethod
    def _get_utm_polygons(self):
        raise NotImplementedError

    @staticmethod
    def _get_utm_from_props(utm_dict):
        """ Return the UTM CRS corresponding to the UTM described by the properties dictionary

        :param utm_dict: Dictionary reporting name of the UTM zone and MGRS grid
        :type utm_dict: dict
        :return: UTM coordinate reference system
        :rtype: sentinelhub.CRS
        """
        return CRS('32{}{}'.format(6 if utm_dict['direction'] == 'N' else 7, str(utm_dict['zone']).zfill(2)))

    def _align_bbox_to_size(self, bbox):
        """ Align input bbox coordinates to be multiples of the bbox size

        :param bbox: Bounding box in UTM coordinates
        :type bbox: sentinelhub.BBox
        :return: BBox objects with coordinates multiples of the bbox size
        :rtype: sentinelhub.BBox
        """
        size_x, size_y = self.bbox_size
        lower_left_x, lower_left_y = bbox.lower_left
        return BBox([(math.floor(lower_left_x / size_x) * size_x, math.floor(lower_left_y / size_y) * size_y),
                     bbox.upper_right], crs=bbox.crs)

    def _make_split(self):
        """ Split each UTM grid into equally sized bboxes in correct UTM zone
        """
        size_x, size_y = self.bbox_size
        self.bbox_list = []
        self.info_list = []

        index = 0

        for utm_cell in self.utm_grid:
            utm_cell_geom, utm_cell_prop = utm_cell
            # the UTM MGRS grid definition contains four 0 zones at the poles (0A, 0B, 0Y, 0Z)
            if utm_cell_prop['zone'] == 0:
                continue
            utm_crs = self._get_utm_from_props(utm_cell_prop)

            intersection = utm_cell_geom.intersection(self.shape_geometry.geometry)

            if not intersection.is_empty and isinstance(intersection, GeometryCollection):
                intersection = MultiPolygon(geo_object for geo_object in intersection
                                            if isinstance(geo_object, (Polygon, MultiPolygon)))

            if not intersection.is_empty:
                intersection = Geometry(intersection, CRS.WGS84).transform(utm_crs)

                bbox_partition = self._align_bbox_to_size(intersection.bbox).get_partition(size_x=size_x, size_y=size_y)

                columns, rows = len(bbox_partition), len(bbox_partition[0])
                for i, j in itertools.product(range(columns), range(rows)):
                    if bbox_partition[i][j].geometry.intersects(intersection.geometry):
                        self.bbox_list.append(bbox_partition[i][j])
                        self.info_list.append(dict(crs=utm_crs.name,
                                                   utm_zone=str(utm_cell_prop['zone']).zfill(2),
                                                   utm_row=utm_cell_prop['row'],
                                                   direction=utm_cell_prop['direction'],
                                                   index=index,
                                                   index_x=i,
                                                   index_y=j))
                        index += 1

    def get_bbox_list(self, buffer=None):
        """ Get list of bounding boxes.

        The CRS is fixed to the computed UTM CRS. This BBox splitter does not support reducing size of output
        bounding boxes

        :param buffer: A percentage of each BBox size increase. This will cause neighbouring bounding boxes to overlap.
        :type buffer: float or None
        :return: List of bounding boxes
        :rtype: list(BBox)
        """
        return super().get_bbox_list(buffer=buffer)


class UtmGridSplitter(BaseUtmSplitter):
    """ Splitter that returns bounding boxes of fixed size aligned to the UTM MGRS grid
    """
    def _get_utm_polygons(self):
        """ Find UTM grid zones overlapping with input area shape

        :return: List of geometries and properties of UTM grid zones overlapping with input area shape
        :rtype: list
        """
        # file downloaded from faculty.baruch.cuny.edu/geoportal/data/esri/world/utmzone.zip
        utm_grid_filename = os.path.join(os.path.dirname(__file__), '.utmzones.geojson')

        if not os.path.isfile(utm_grid_filename):
            raise IOError('UTM grid definition file does not exist: %s' % os.path.abspath(utm_grid_filename))

        with open(utm_grid_filename) as utm_grid_file:
            utm_grid = json.load(utm_grid_file)['features']

        utm_geom_list = [shapely.geometry.shape(utm_zone['geometry']) for utm_zone in utm_grid]
        utm_prop_list = [dict(zone=utm_zone['properties']['ZONE'],
                              row=utm_zone['properties']['ROW_'],
                              direction='N' if utm_zone['properties']['ROW_'] >= 'N' else 'S') for utm_zone in utm_grid]

        return list(zip(utm_geom_list, utm_prop_list))


class UtmZoneSplitter(BaseUtmSplitter):
    """ Splitter that returns bounding boxes of fixed size aligned to the equator and the UTM zones.
    """
    LNG_MIN, LNG_MAX, LNG_UTM = -180, 180, 6
    LAT_MIN, LAT_MAX, LAT_EQ = -80, 84, 0

    def _get_utm_polygons(self):
        """ Find UTM zones overlapping with input area shape

        The returned geometry corresponds to the a triangle ranging from the equator to the north/south pole

        :return: List of geometries and properties of UTM zones overlapping with input area shape
        :rtype: list
        """
        utm_geom_list = []
        for lat in [(self.LAT_EQ, self.LAT_MAX), (self.LAT_MIN, self.LAT_EQ)]:
            for lng in range(self.LNG_MIN, self.LNG_MAX, self.LNG_UTM):
                points = []
                # A new point is added per each degree - this is inline with geometries used by UtmGridSplitter
                # In the future the number of points will be calculated according to bbox_size parameter
                for degree in range(lat[0], lat[1]):
                    points.append((lng, degree))
                for degree in range(lng, lng + self.LNG_UTM):
                    points.append((degree, lat[1]))
                for degree in range(lat[1], lat[0], -1):
                    points.append((lng + self.LNG_UTM, degree))
                for degree in range(lng + self.LNG_UTM, lng, -1):
                    points.append((degree, lat[0]))

                utm_geom_list.append(Polygon(points))

        utm_prop_list = [dict(zone=zone, row='', direction=direction)
                         for direction in ['N', 'S'] for zone in range(1, 61)]

        return list(zip(utm_geom_list, utm_prop_list))
